from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Literal

import numpy
from ewokscore import Task
from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import MissingData
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict
from silx.io.dictdump import dicttonx
from silx.math.combo import min_max

from darfix import dtypes
from darfix.core.moment_types import MomentType

from ..core.grainplot import DimensionRange
from ..core.grainplot import GrainPlotMaps
from ..core.grainplot import OrientationDistData
from ..core.grainplot import compute_mosaicity
from ..core.grainplot import generate_grain_maps_nxdict

_logger = logging.getLogger(__file__)


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: dtypes.Dataset
    """ Input dataset containing a stack of images """
    dimensions: tuple[int, int] = (0, 1)
    """Dimension indices to use for the maps. Default is (0, 1), which means the two first dimensions."""
    range: tuple[DimensionRange | None, DimensionRange | None] = (None, None)
    """Dimensionrange for the two dimensions. If None, use the Center of Mass min and max for the both dimensions."""
    save_maps: bool = True
    """Whether to save the maps to file. Default is True."""
    filename: str | MissingData = MISSING_DATA
    """Only used if save_maps is True. Filename to save the maps to. Default is 'maps.h5' in the dataset directory."""
    orientation_img_origin: Literal["dims", "center"] = "dims"
    "Origin for the orientation distribution image. Can be 'dims', 'center' or None. Default is 'dims'."


class GrainPlot(
    Task,
    input_model=Inputs,
    output_names=["dataset"],
):
    """Generates and saves maps of Center of Mass, FWHM, Skewness, Kurtosis, Orientation distribution and Mosaicity."""

    def run(self):

        inputs = Inputs(**self.get_input_values())

        default_filename = os.path.join(inputs.dataset.dataset._dir, "maps.h5")
        filename: str = self.get_input_value("filename", default_filename)

        moments = inputs.dataset.dataset.apply_moments()
        grainPlotMaps = GrainPlotMaps.from_dataset(inputs.dataset)

        # mosaicity and orientation can only be computed for 2D+ datasets
        if grainPlotMaps.dims.ndim > 1 and inputs.save_maps:
            dimension1, dimension2 = inputs.dimensions
            dimension1_range, dimension2_range = inputs.range

            if dimension1_range is None:
                dimension1_range = _computeMinMax(moments[dimension1][MomentType.COM])
            if dimension2_range is None:
                dimension2_range = _computeMinMax(moments[dimension2][MomentType.COM])

            mosaicity = compute_mosaicity(
                moments,
                x_dimension=dimension1,
                y_dimension=dimension2,
                x_dimension_range=dimension1_range,
                y_dimension_range=dimension2_range,
            )

            orientation_dist_data = OrientationDistData(
                grainPlotMaps,
                x_dimension=dimension1,
                y_dimension=dimension2,
                x_dimension_range=dimension1_range,
                y_dimension_range=dimension2_range,
            )
            assert orientation_dist_data is not None
        else:
            orientation_dist_data = None
            mosaicity = None

        # Save data if asked
        if inputs.save_maps:
            nxdict = generate_grain_maps_nxdict(
                grainPlotMaps, mosaicity, orientation_dist_data
            )
            os.makedirs(Path(filename).parent, exist_ok=True)
            dicttonx(nxdict, filename)

        self.outputs.dataset = inputs.dataset


def _computeMinMax(array: numpy.ndarray) -> DimensionRange:
    min_max_result = min_max(array)
    return min_max_result.minimum, min_max_result.maximum



from __future__ import annotations

from silx.gui import qt

from darfix import dtypes
from darfix.core.utils import OperationAborted
from darfix.gui.utils.message import missing_dataset_msg
from darfix.gui.utils.standardButtonBox import StandardButtonBox
from orangecontrib.darfix.widgets.datasetWidgetBase import DatasetWidgetBase


class OperationWidgetBase(DatasetWidgetBase, openclass=True):
    """
    Base class for an Orange Ewoks Widget that can run one or more operations
    and that manage a 'Reset' and an 'Abort' button.

    Note that the 'Abort' button will work only if ewoks task implement
    the `cancel` abstract method.
    """

    want_main_area = True
    want_control_area = False

    sigOperationApplied = qt.Signal(dtypes.Dataset)

    def __init__(
        self,
        buttonTypes: qt.QDialogButtonBox.StandardButton = qt.QDialogButtonBox.StandardButton.NoButton,
    ):
        super().__init__()

        # buttons
        self.__buttons = StandardButtonBox(parent=self, additionalButtons=buttonTypes)
        self.__buttons.okButton.setEnabled(False)

        self.__widget = self.createMainWidget(self.get_default_input_values())
        self.mainArea.layout().addWidget(self.__widget)
        self.layout().addWidget(self.__buttons)

        self.set_dynamic_input("copy_dataset", True)

        self.__saveAction = self.getDatasetMenu().addAction("Save")
        self.__saveAction.setDisabled(True)
        self.__saveAction.setShortcut("Ctrl+Shift+S")

        # connect signal / slot
        self.__buttons.applyButton.clicked.connect(self.__onApplyClicked)
        self.__buttons.okButton.clicked.connect(self.__onOkClicked)
        self.__buttons.resetButton.clicked.connect(self.__onResetClicked)
        self.__buttons.abortButton.clicked.connect(self.__onAbortClicked)
        self.__saveAction.triggered.connect(self.__onSaveDataset)
        self.datasetChanged.connect(self.__onDatasetChanged)

    @property
    def buttons(self) -> StandardButtonBox:
        return self.__buttons

    @property
    def mainWidget(self) -> qt.QWidget:
        return self.__widget

    @property
    def __backupDataset(self) -> dtypes.Dataset | None:
        return self.get_task_input_value("dataset", None)

    def saveMainWidget(self) -> None:
        """
        inherited class need to override this method.

        Save parameters as default inputs with `set_default_input` here.
        """
        raise NotImplementedError("This is an abstract method.")

    def createMainWidget(self, default_inputs: dict) -> qt.QWidget:
        """
        inherited class need to override this method.

        Instantiate and setup main widget here.

        :param default_inputs: Dictionary containing the default inputs that are set.
        :return : An instance of the created main widget
        """
        raise NotImplementedError("This is an abstract method.")

    def __onOkClicked(self) -> None:
        if len(self.get_task_outputs()) == 0:
            # if no output at all this is an unexpected behaviour
            raise RuntimeError("Cannot go to next step because outputs are empty.")
        self.close()
        self.propagate_downstream()

    def __onResetClicked(self):
        self.reset()

    def __onAbortClicked(self):
        self.cancel_running_task()

    def __onApplyClicked(self):
        self.__buttons.applyButton.setDisabled(True)
        self.saveMainWidget()
        self.__buttons.setIsComputing(True)
        self.execute_ewoks_task_without_propagation()

    def __onSaveDataset(self):
        dataset = self.get_task_output_value("dataset", None)
        # All exception below are unexpected as save action is disabled if no successful output
        assert dataset is not None, "output dataset should not be None"
        assert isinstance(
            dataset, dtypes.Dataset
        ), f"Dataset bad type : {type(dataset)}"
        assert dataset.dataset is not None, "output image stack should not be None"
        imgDataset = dataset.dataset
        if imgDataset is not None:
            filename, _ = qt.QFileDialog().getSaveFileName(
                None,
                "Save Darfix dataset",
                filter="Darfix HDF5 file (*.h5)",
            )
            if filename:
                imgDataset.save(filename)
        else:
            missing_dataset_msg()

    def __onDatasetChanged(self):
        self.__buttons.okButton.setDisabled(True)
        self.__buttons.applyButton.setEnabled(True)
        self.__saveAction.setDisabled(True)

    def reset(self):
        self.setDataset(self.__backupDataset)

    def task_output_changed(self) -> None:
        self.__buttons.setIsComputing(False)
        if self.task_succeeded:
            self.__buttons.okButton.setEnabled(True)

            output = self.get_task_output_value("dataset", None)
            if output is None:
                raise RuntimeError(
                    "Unexpected error : Task succeeded but output is None"
                )
            if output is self.get_task_input_value("dataset", None):
                raise RuntimeError("Unexpected error : Input and Output are the same.")

            self.__saveAction.setEnabled(True)

            self.sigOperationApplied.emit(output)
        else:
            self.__buttons.applyButton.setEnabled(True)

    def isAborted(self) -> bool:
        return isinstance(self.task_exception, OperationAborted)

    def closeEvent(self, evt) -> None:
        super().closeEvent(evt)
        self.saveMainWidget()



from __future__ import annotations

import numpy
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot.items import ImageBase
from silx.image.marchingsquares import find_contours

from darfix.gui.grainplot.dimensionRangeSlider2D import DimensionRangeSlider2D

from ...core.grainplot import GrainPlotMaps
from ...core.grainplot import MultiDimMomentType
from ...core.grainplot import OrientationDistData
from ...core.grainplot import compute_mosaicity
from ...core.moment_types import MomentType
from .flashlight import PlotWithFlashlight
from .oridist_toolbar import OriDistButtonIds
from .oridist_toolbar import OriDistToolbar
from .utils import MapType
from .utils import add_image_with_transformation


class MosaicityWidget(qt.QWidget):
    """Widget to display and explore mosaicity plots"""

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent)

        self._grainPlotMaps: GrainPlotMaps | None = None
        self._orientation_dist_data: OrientationDistData | None = None
        self._mosaicity: numpy.ndarray | None = None
        self.contours = {}

        layout = qt.QGridLayout()

        #
        # Mosaicity plot (left) and Orientation Distribution (Ori Dist) plot (Right)
        # Row 0
        #

        self._mosaicityPlot = PlotWithFlashlight(parent=self)
        self._mosaicityPlot.setKeepDataAspectRatio(True)

        layout.addWidget(self._mosaicityPlot, 0, 0)

        self._oriDistPlot = PlotWithFlashlight(parent=self)
        self._oriDistPlot.setKeepDataAspectRatio(True)

        layout.addWidget(self._oriDistPlot, 0, 1)

        self._mosaicityPlot.sigFlashlightMove.connect(self._onMoveInMosaPlot)
        self._mosaicityPlot.sigFlashlightHide.connect(self._oriDistPlot.clearPlotMask)

        self._oriDistPlot.sigFlashlightMove.connect(
            self._onMoveInOrientationDistribPlot
        )
        self._oriDistPlot.sigFlashlightHide.connect(self._mosaicityPlot.clearPlotMask)

        #
        # Orientation Distribution (Ori Dist) Configuration dock widget as an action
        # in the `_oriDistPlot` tool bar.`
        #

        self._oriDistToolbar = OriDistToolbar()

        self._oriDistPlot.addToolBar(self._oriDistToolbar)

        self._oriDistToolbar.sigContourChanged.connect(self._computeContours)
        self._oriDistToolbar.sigReplot.connect(self._plotOrientationData)

        #
        # info label then separator
        # Row 1 and 2
        #
        infoLayout = qt.QHBoxLayout()
        infoLayout.addWidget(qt.QLabel("Dimension values:"))
        self._infosLabel1 = qt.QLabel()
        self._infosLabel2 = qt.QLabel()
        infoLayout.addWidget(self._infosLabel1)
        infoLayout.addWidget(self._infosLabel2)
        layout.addLayout(infoLayout, 1, 0, 1, 2)

        separator = qt.QFrame()
        separator.setFrameShape(qt.QFrame.HLine)
        separator.setFrameShadow(qt.QFrame.Sunken)
        layout.addWidget(separator, 2, 0, 1, 2)

        #
        # zsum threshold
        # Row 3
        #

        thresholdLayout = qt.QHBoxLayout()

        self._zsum: None | numpy.ndarray = None
        """Keep zsum to avoid compute each time we update plots"""
        self._thresholdSB = qt.QSpinBox()
        """ Discard mosa values below this threshold """
        self._thresholdSB.setButtonSymbols(qt.QAbstractSpinBox.ButtonSymbols.NoButtons)
        self._thresholdSB.editingFinished.connect(self._computeMosaicityAndOriDist)
        thresholdLayout.addWidget(qt.QLabel("ZSum Threshold : "))
        thresholdLayout.addWidget(self._thresholdSB)
        thresholdLayout.addStretch(1)
        layout.addLayout(thresholdLayout, 3, 0)

        #
        # Dimension range sliders
        # Row 4
        #

        self._sliders = DimensionRangeSlider2D()
        self._sliders.sigChanged.connect(self._computeMosaicityAndOriDist)
        layout.addWidget(self._sliders, 4, 0, 1, 2)

        self.setLayout(layout)

    @property
    def dimension1(self) -> int:
        return self._sliders.indexDimX()

    @property
    def dimension2(self) -> int:
        return self._sliders.indexDimY()

    def getMosaicity(self):
        return self._mosaicity

    def getContoursImage(self):
        return self._oriDistPlot.getImage()

    def getOrientationDist(self):
        return self._orientation_dist_data

    def setGrainPlotMaps(self, grainPlotMaps: GrainPlotMaps):
        self._grainPlotMaps = grainPlotMaps
        self._thresholdSB.setMaximum(int(self._grainPlotMaps.zsum.max()))

        self._mosaicityPlot.setGraphTitle(
            self._grainPlotMaps.title + "\n" + MapType.MOSAICITY.value
        )
        self._oriDistPlot.setGraphTitle(
            self._grainPlotMaps.title + "\n" + MultiDimMomentType.ORIENTATION_DIST.value
        )

        self._sliders.setDimensions(self._grainPlotMaps.dims)

        self._computeMosaicityAndOriDist()

    def _plotMosaicity(self):
        if self._grainPlotMaps is None:
            return

        if self._mosaicity is not None:
            add_image_with_transformation(
                self._mosaicityPlot, self._mosaicity, self._grainPlotMaps.transformation
            )

    def _plotOrientationData(self, state=None):
        if self._grainPlotMaps is None or self._orientation_dist_data is None:
            return

        self._oriDistPlot.resetZoom()

        origin = self._orientation_dist_data.origin(
            self._oriDistToolbar.getOriginType()
        )

        self._oriDistPlot.addImage(
            self._orientation_dist_data.rgb_key,
            legend="Color key",
            xlabel=self._orientation_dist_data.x_label,
            ylabel=self._orientation_dist_data.y_label,
            origin=origin,
            scale=self._orientation_dist_data.rgb_key_plot_scale(),
        )
        maxOpacityValue = self._oriDistToolbar.getColorKeyOpacityValue()
        colors = numpy.zeros(shape=(maxOpacityValue, 4), dtype=numpy.uint8)
        colors[..., -1] = numpy.arange(maxOpacityValue - 1, -1, -1)

        self._oriDistPlot.addImage(
            (
                self._orientation_dist_data.smooth_data
                if self._oriDistToolbar.medianFilterEnabled()
                else self._orientation_dist_data.data
            ),
            legend="Data",
            xlabel=self._orientation_dist_data.x_label,
            ylabel=self._orientation_dist_data.y_label,
            origin=origin,
            colormap=Colormap(
                colors=colors,
                normalization=self._oriDistToolbar.getNormalization(),
                vmax=(
                    1
                    if self._oriDistToolbar.getDataPlotType()
                    == OriDistButtonIds.COLOR_KEY
                    else None
                ),
            ),
            scale=self._orientation_dist_data.data_plot_scale(),
        )
        if self._oriDistToolbar.getDataPlotType() == OriDistButtonIds.COLOR_KEY:
            self._oriDistPlot.setActiveImage("Color key")
        else:
            self._oriDistPlot.setActiveImage("Data")

    def _computeContours(self):
        """
        Compute contours map based on orientation distribution.
        """
        self._oriDistPlot.remove(kind="curve")

        if not self._oriDistToolbar.isContourEnabled():
            return

        orientationImagePlot: ImageBase | None = self._oriDistPlot.getImage()

        if self._orientation_dist_data is None or orientationImagePlot is None:
            return

        min_orientation = numpy.min(self._orientation_dist_data.smooth_data)
        max_orientation = numpy.max(self._orientation_dist_data.smooth_data)

        polygons = []
        levels = []

        normalization = self._oriDistToolbar.getNormalization()

        if normalization == "log":
            isoValues = numpy.logspace(
                numpy.log10(max(min_orientation, 1)),
                numpy.log10(max_orientation),
                int(self._oriDistToolbar.getLevelCount()),
            )
        elif normalization == "linear":
            isoValues = numpy.linspace(
                min_orientation,
                max_orientation,
                int(self._oriDistToolbar.getLevelCount()),
            )
        else:
            raise ValueError(f"Unhandled normalization value `{normalization}`")
        for isoValue in isoValues:
            polygons.append(
                find_contours(self._orientation_dist_data.smooth_data, isoValue)
            )
            levels.append(isoValue)

        colormap = Colormap(
            name="temperature",
            vmin=min_orientation,
            vmax=max_orientation,
            normalization=normalization,
        )
        colors = colormap.applyToData(levels)
        self.contours = {}
        for ipolygon, polygon in enumerate(polygons):
            # iso contours
            for icontour, contour in enumerate(polygon):
                if len(contour) == 0:
                    continue

                x = contour[:, 1]
                y = contour[:, 0]
                x_pts, y_pts = self._orientation_dist_data.to_motor_coordinates(
                    x, y, self._oriDistToolbar.getOriginType()
                )
                legend = f"poly{icontour}.{ipolygon}"
                self.contours[legend] = {
                    "points": (x_pts.copy(), y_pts.copy()),
                    "color": colors[ipolygon],
                    "value": levels[ipolygon],
                    "pixels": (x, y),
                }
                self._oriDistPlot.addCurve(
                    x=x_pts,
                    y=y_pts,
                    linestyle="-",
                    linewidth=2.0,
                    legend=legend,
                    resetzoom=False,
                    color=colors[ipolygon],
                )

    def _updatePositionInfoLabel(self, firstDimValue: float, secondDimValue: float):
        dim1Name = self._grainPlotMaps.dims.get(self.dimension1).name
        dim2Name = self._grainPlotMaps.dims.get(self.dimension2).name
        self._infosLabel1.setText(f"{dim1Name} = {firstDimValue:.4}")
        self._infosLabel2.setText(f"{dim2Name} = {secondDimValue:.4}")

    def _onMoveInMosaPlot(self):

        slices = self._mosaicityPlot.getRoi().getSlices()
        i, j = self._mosaicityPlot.getRoi().getCenterIJ()

        com_x_range = self._grainPlotMaps.moments_dims[self.dimension1][MomentType.COM][
            slices
        ].ravel()
        com_y_range = self._grainPlotMaps.moments_dims[self.dimension2][MomentType.COM][
            slices
        ].ravel()
        com_x_range = com_x_range[numpy.isfinite(com_x_range)]
        com_y_range = com_y_range[numpy.isfinite(com_y_range)]

        self._updatePositionInfoLabel(
            self._grainPlotMaps.moments_dims[self.dimension1][MomentType.COM][j, i],
            self._grainPlotMaps.moments_dims[self.dimension2][MomentType.COM][j, i],
        )

        currentImg = self._oriDistPlot.getActiveImage()

        scale_x, scale_y = currentImg.getScale()
        origin_x, origin_y = currentImg.getOrigin()
        indices_i = ((com_x_range - origin_x) / scale_x).astype(int)
        indices_j = ((com_y_range - origin_y) / scale_y).astype(int)
        indices_ji = numpy.stack([indices_j, indices_i])

        shape = self._oriDistPlot.getPlotMaskShape()

        mask = numpy.ones(shape, dtype=bool)
        discardOutOfBoundsMask = numpy.all(
            (indices_ji >= 0)
            & (indices_ji < numpy.asarray(mask.shape)[:, numpy.newaxis]),
            axis=0,
        )
        indices_ji = indices_ji[:, discardOutOfBoundsMask]
        indices_j, indices_i = indices_ji

        mask[indices_j, indices_i] = False
        self._oriDistPlot.createPlotMask(mask)

    def _onMoveInOrientationDistribPlot(self):

        x, y = self._oriDistPlot.getRoi().getCenter()

        self._updatePositionInfoLabel(x, y)

        com_x = self._grainPlotMaps.moments_dims[self.dimension1][MomentType.COM]
        com_y = self._grainPlotMaps.moments_dims[self.dimension2][MomentType.COM]

        w, h = self._oriDistPlot.getRoi().getSize() / 2

        iso_x = numpy.isclose(com_x, x, atol=w)
        iso_y = numpy.isclose(com_y, y, atol=h)
        mask = ~numpy.all(numpy.stack((iso_x, iso_y)), axis=0)
        self._mosaicityPlot.createPlotMask(mask)

    def _computeMosaicityAndOriDist(self):
        """
        Compute mosaicity and orientation distribution.

        Called when dimensions are changed (by setting the dataset or user interaction).
        """

        if self._grainPlotMaps is None:
            return

        maskZsum = self._grainPlotMaps.zsum <= self._thresholdSB.value()

        zsum = self._grainPlotMaps.zsum.copy()
        zsum[maskZsum] = 0

        self._mosaicity = compute_mosaicity(
            self._grainPlotMaps.moments_dims,
            x_dimension=self.dimension1,
            y_dimension=self.dimension2,
            x_dimension_range=self._sliders.rangeDimX(),
            y_dimension_range=self._sliders.rangeDimY(),
        )
        self._mosaicity[maskZsum] = 0, 0, 0  # Black

        self._orientation_dist_data = OrientationDistData(
            self._grainPlotMaps,
            x_dimension=self.dimension1,
            y_dimension=self.dimension2,
            x_dimension_range=self._sliders.rangeDimX(),
            y_dimension_range=self._sliders.rangeDimY(),
            zsum=zsum,
        )
        self._plotMosaicity()
        self._plotOrientationData()
        self._computeContours()



from __future__ import annotations

import numpy
from ewoksorange.gui.parameterform import block_signals
from silx.gui import qt
from silx.gui.widgets.RangeSlider import RangeSlider


class RangeSliderWithSpinBox(qt.QWidget):
    """RangeSlider with spin boxes to select a numeric range."""

    sigValueChanged = qt.Signal(float, float)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.__spinBoxMin = qt.QDoubleSpinBox()
        self.__spinBoxMax = qt.QDoubleSpinBox()
        self.__slider = RangeSlider()

        for slider in self.__spinBoxMin, self.__spinBoxMax:
            slider.setButtonSymbols(qt.QAbstractSpinBox.ButtonSymbols.NoButtons)
            slider.setRange(-numpy.inf, numpy.inf)
            slider.setDecimals(3)

        layout = qt.QHBoxLayout(self)

        layout.addWidget(self.__spinBoxMin)
        layout.addWidget(self.__slider, 1)
        layout.addWidget(self.__spinBoxMax)

        self.__slider.sigValueChanged.connect(self.sigValueChanged)
        self.__slider.sigValueChanged.connect(self._onSliderUpdate)
        self.__spinBoxMin.editingFinished.connect(self._onSpinMinEdited)
        self.__spinBoxMax.editingFinished.connect(self._onSpinMaxEdited)

    def setRange(self, vmin: float, vmax: float):
        if vmax <= vmin:
            raise ValueError(f"max {vmax} <= min {vmin}")

        with block_signals(self.__spinBoxMin):
            self.__spinBoxMin.setMaximum(vmax)
            self.__spinBoxMin.setValue(vmin)

        with block_signals(self.__spinBoxMax):
            self.__spinBoxMax.setMinimum(vmin)
            self.__spinBoxMax.setValue(vmax)

        with block_signals(self.__slider):
            self.__slider.setRange(vmin, vmax)
            self.__slider.setValues(vmin, vmax)

    def getValues(self) -> tuple[float, float]:
        """Return the current selected (min, max) values."""
        return self.__slider.getValues()

    def getRange(self) -> tuple[float, float]:
        return self.__slider.getRange()

    def _onSpinMinEdited(self):
        spinMinValue = self.__spinBoxMin.value()

        # Minimum allowed value of `__spinMax` is `__spinBoxMin` value
        self.__spinBoxMax.setMinimum(spinMinValue)

        with block_signals(self.__slider):
            # Update slider minimum if spinbox min value is below range min
            if spinMinValue < self.__slider.getMinimum():
                self.__slider.setMinimum(spinMinValue)
            # update slider values
            self.__slider.setValues(spinMinValue, self.__spinBoxMax.value())

        self.sigValueChanged.emit(spinMinValue, self.__spinBoxMax.value())

    def _onSpinMaxEdited(self):
        spinMaxValue = self.__spinBoxMax.value()

        # Maximum allowed value of `__spinBoxMin` is `__spinBoxMax` value
        self.__spinBoxMin.setMaximum(spinMaxValue)

        with block_signals(self.__slider):
            # Update slider maximum if spinbox max value is above range max
            if spinMaxValue > self.__slider.getMaximum():
                self.__slider.setMaximum(spinMaxValue)
            # update slider values
            self.__slider.setValues(self.__spinBoxMin.value(), spinMaxValue)

        self.sigValueChanged.emit(self.__spinBoxMin.value(), spinMaxValue)

    def _onSliderUpdate(self):
        v0, v1 = self.__slider.getValues()
        self.__spinBoxMin.setValue(v0)
        self.__spinBoxMax.setValue(v1)



import numpy
from silx.gui import qt


def createCustomDoubleSpinBox(initialValue: float = 0.0) -> qt.QDoubleSpinBox:
    """
    Custom double spin box with :
    - No button
    - No min / max limits
    - 4 decimals
    - an optional `initialValue` (0.0 by default)
    """
    spinBox = qt.QDoubleSpinBox()
    spinBox.setValue(initialValue)
    spinBox.setDecimals(4)
    spinBox.setMaximum(numpy.inf)
    spinBox.setMinimum(-numpy.inf)
    spinBox.setButtonSymbols(qt.QAbstractSpinBox.ButtonSymbols.NoButtons)
    return spinBox



from __future__ import annotations

from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread
from silx.gui import qt

from darfix import dtypes
from darfix.core.dataset import ImageDataset
from darfix.gui.utils.message import show_error_msg


class DatasetWidgetBase(OWEwoksWidgetOneThread, openclass=True):
    """
    Orange Ewoks Widget with a "dataset" dynamic input.

    Base class manage

    - mainArea enable / disable when a task is running.
    - load / save dataset menu
    - `setDataset()` public method
    - `datasetChanged` signal : Children should use this signal instead of `handleNewSignals`
    """

    datasetChanged = qt.Signal(dtypes.Dataset)

    def __init__(self):
        super().__init__()
        self.task_executor.finished.connect(self.__onFinishTask)
        self.task_executor.started.connect(self.__onStartEwoksTask)

        self.__datasetMenu: qt.QMenu = self.menuBar().addMenu("&Dataset")

        loadAction = self.__datasetMenu.addAction("Load")
        loadAction.setShortcut("Ctrl+O")

        loadAction.triggered.connect(self._onLoadDataset)

    def handleNewSignals(self) -> None:
        self._updateDataset(True)
        # super().handleNewSignals() is not called because we do not want to trigger workflow

    def setDataset(self, dataset: dtypes.Dataset | None, pop_up: bool = False) -> None:
        self.set_dynamic_input("dataset", dataset)
        self._updateDataset(pop_up)

    def _updateDataset(self, pop_up: bool) -> None:
        dataset = self._getDatasetTaskInput()
        if dataset is None:
            return
        if not isinstance(dataset, dtypes.Dataset):
            show_error_msg(f"Bad dataset type : {type(dataset)}")
            return
        self.datasetChanged.emit(dataset)
        if pop_up:
            self.open()

    def _getDatasetTaskInput(self) -> dtypes.Dataset | None:
        return self.get_task_input_value("dataset", None)

    def _onLoadDataset(self):
        filename, _ = qt.QFileDialog().getOpenFileName(
            None, "Open Darfix dataset save", filter="Darfix HDF5 file (*.h5)"
        )
        if filename:
            dataset = dtypes.Dataset(ImageDataset.load(filename))
            self.setDataset(dataset)

    def __onFinishTask(self):
        if self.mainArea is not None:
            self.mainArea.setEnabled(True)

    def __onStartEwoksTask(self):
        if self.mainArea is not None:
            self.mainArea.setDisabled(True)

    def getDatasetMenu(self) -> qt.QMenu:
        return self.__datasetMenu



from __future__ import annotations

from enum import Enum

from silx.gui import icons
from silx.gui import qt

from ...dtypes import AxisType
from ._oridist_toolbar_buttons import _AxisTypeToolButton
from ._oridist_toolbar_buttons import _ContourConfigToolButton
from ._oridist_toolbar_buttons import _DatatypeToolbutton
from ._oridist_toolbar_buttons import _NormalizationToolButton


class OriDistButtonIds(Enum):
    DATA = 0
    """Data of the orientation distribution histogram 2D"""
    COLOR_KEY = 1
    """ Data as a HSV color map"""


class OriDistToolbar(qt.QToolBar):
    """
    Custom toolbar for the Orientation Distribution plot configuration
    """

    sigContourChanged = qt.Signal()
    sigReplot = qt.Signal()

    def __init__(self):

        super().__init__()

        self.__contourToolButton = _ContourConfigToolButton()

        self.__scaleToolButton = _NormalizationToolButton()

        self.__dataPlotTypeToolButton = _DatatypeToolbutton()

        self.__axisToolButton = _AxisTypeToolButton()

        self.__medianFilterAction = qt.QAction()
        self.__medianFilterAction.setCheckable(True)
        self.__medianFilterAction.setChecked(True)
        self.__medianFilterAction.setIcon(
            icons.getQIcon("darfix:gui/icons/median-filter")
        )
        self.__medianFilterAction.setToolTip("Apply a 2D median filter")

        self.addWidget(self.__scaleToolButton)
        self.addWidget(self.__axisToolButton)
        self.addWidget(self.__dataPlotTypeToolButton)
        self.addWidget(self.__contourToolButton)
        self.addAction(self.__medianFilterAction)

        # connect signal / slot
        self.__contourToolButton.sigChanged.connect(self.sigContourChanged)
        self.__scaleToolButton.sigChanged.connect(self.sigReplot)
        self.__scaleToolButton.sigChanged.connect(self.sigContourChanged)
        self.__axisToolButton.sigChanged.connect(self.sigReplot)
        self.__axisToolButton.sigChanged.connect(self.sigContourChanged)
        self.__dataPlotTypeToolButton.sigChanged.connect(self.sigReplot)
        self.__medianFilterAction.toggled.connect(self.sigReplot)

    def getOriginType(self) -> AxisType:
        return self.__axisToolButton.getCurrentAxisType()

    def getNormalization(self) -> str:
        return self.__scaleToolButton.getNormalization()

    def getLevelCount(self) -> int:
        return self.__contourToolButton.value()

    def isContourEnabled(self) -> bool:
        return self.__contourToolButton.isChecked()

    def getColorKeyOpacityValue(self) -> int:
        return 256 - self.__dataPlotTypeToolButton.value()

    def getDataPlotType(self) -> OriDistButtonIds:
        if self.__dataPlotTypeToolButton.isChecked():
            return OriDistButtonIds.COLOR_KEY
        else:
            return OriDistButtonIds.DATA

    def medianFilterEnabled(self) -> bool:
        return self.__medianFilterAction.isChecked()



from silx.gui.plot.actions import PlotAction


class FlashlightModeAction(PlotAction):
    """QAction controlling the the Flashlight Mode to highlight a portion of the mosaicity.

    :param plot: :class:`.PlotWidget` instance on which to operate
    :param parent: See :class:`QAction`
    """

    INTERACTION_MODE = "select"

    def __init__(self, plot, parent=None):
        super().__init__(
            plot,
            icon="darfix:gui/icons/flashlight",
            text="Flashlight mode",
            tooltip="Highlight a portion of the mosaicity",
            triggered=self._actionTriggered,
            checkable=True,
            parent=parent,
        )
        # Listen to mode change
        self.plot.sigInteractiveModeChanged.connect(self._modeChanged)
        # Init the state
        self._modeChanged(None)

    def _modeChanged(self, source):
        modeDict = self.plot.getInteractiveMode()
        self.setChecked(modeDict["mode"] == self.INTERACTION_MODE)

    def _actionTriggered(self, checked=False):
        plot = self.plot
        if plot is not None:
            plot.setInteractiveMode(self.INTERACTION_MODE, source=self)



from __future__ import annotations

from functools import partial

from ewoksorange.gui.parameterform import block_signals
from silx.gui import qt

from darfix.core.dimension import AcquisitionDims
from darfix.core.dimension import Dimension
from darfix.gui.utils.rangeSlider import RangeSliderWithSpinBox


class _DimensionRangeSlider(qt.QWidget):
    """
    RangeSlider for a darfix dimension
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.dimCB = qt.QComboBox()
        self.range = RangeSliderWithSpinBox()

        layout = qt.QHBoxLayout()
        layout.addWidget(self.dimCB)
        layout.addWidget(self.range)
        self.setLayout(layout)

    def setCurrentDimension(self, dims: AcquisitionDims, index: int):
        dim = dims.get(index)
        self.setDimensionRange(dim)
        with block_signals(self.dimCB):
            self.dimCB.clear()
            self.dimCB.addItems(dims.get_names())
            self.dimCB.setCurrentIndex(index)

    def setDimensionRange(self, dim: Dimension):
        self.range.setRange(dim.min(), dim.max())

    def currentDimensionIndex(self) -> int:
        return self.dimCB.currentIndex()

    def setDimensionIndex(self, idx: int):
        self.dimCB.setCurrentIndex(idx)


class DimensionRangeSlider2D(qt.QWidget):
    """
    Two dimension sliders and two combo boxes to select dimension in the mosaicity plot
    """

    sigChanged = qt.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self.__dims: None | AcquisitionDims = None

        layout = qt.QHBoxLayout()

        self.__sliders = _DimensionRangeSlider(), _DimensionRangeSlider()

        for slider in self.__sliders:
            layout.addWidget(slider)
            slider.range.sigValueChanged.connect(self.sigChanged)
        self.__sliders[0].dimCB.currentIndexChanged.connect(
            partial(self._onDimensionChanged, self.__sliders[0], self.__sliders[1])
        )
        self.__sliders[1].dimCB.currentIndexChanged.connect(
            partial(self._onDimensionChanged, self.__sliders[1], self.__sliders[0])
        )

        self.setLayout(layout)

    def indexDimX(self) -> int:
        return self.__sliders[0].currentDimensionIndex()

    def indexDimY(self) -> int:
        return self.__sliders[1].currentDimensionIndex()

    def rangeDimX(self) -> tuple[float, float]:
        return self.__sliders[0].range.getValues()

    def rangeDimY(self) -> tuple[float, float]:
        return self.__sliders[1].range.getValues()

    def setDimensions(self, dims: AcquisitionDims):
        self.__dims = dims
        for idx, slider in enumerate(self.__sliders):
            slider.setCurrentDimension(dims, idx)

    def _onDimensionChanged(
        self, slider: _DimensionRangeSlider, other: _DimensionRangeSlider
    ):

        slider.setDimensionRange(self.__dims.get(slider.currentDimensionIndex()))

        if slider.currentDimensionIndex() == other.currentDimensionIndex():
            if other.currentDimensionIndex() == 0:
                other.setDimensionIndex(1)
            else:
                other.setDimensionIndex(0)
            # Signal emitted by the second slider
        else:
            self.sigChanged.emit()



from __future__ import annotations

import logging
import queue

import numpy
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot import Plot2D
from silx.gui.widgets.FrameBrowser import HorizontalSliderWithBrowser

from darfix import config
from darfix import dtypes
from darfix.core.noiseremoval import BackgroundType
from darfix.core.noiseremoval import NoiseRemovalOperation
from darfix.core.noiseremoval import apply_noise_removal_operation
from darfix.core.noiseremoval import create_background_substraction_operation
from darfix.core.state_of_operation import Operation
from darfix.gui.operationThread import OperationThread

from .operation_list_widget import OperationListWidget
from .parameters_widget import ParametersWidget

_logger = logging.getLogger(__name__)


class PreviewThread(qt.QThread):
    """Execute on the fly computation when plot is updated in order to preview the noise removal for that frame"""

    sigDataReady = qt.Signal(numpy.ndarray)
    """
    Emitted when one processed frame data has been computed. This might not be the current frame displayed but the latest computed one.
    Can be emitted several times during one run.
    """

    QUEUE_TIMEOUT = 0.2

    def __init__(self, parent):
        super().__init__(parent)
        self.queue = queue.Queue()

    def _onTheFlyComputation(self):
        """
        Process operations in image and emit `sigDataReady` signals while the queue is NOT empty for at least `QUEUE_TIMEOUT` sec.
        The Queue will continue growing until the user release the slider and the processing is done.

        raise queue.Empty when exit
        """
        while True:
            img, ops = self.queue.get(timeout=self.QUEUE_TIMEOUT)
            if self.queue.qsize() > 1:
                # skip this image : process is too slow
                continue
            for op in ops:
                apply_noise_removal_operation(img, op)
            self.sigDataReady.emit(img)

    def run(self):
        try:
            self._onTheFlyComputation()
        except queue.Empty:
            # Thread will finish as timeout reached
            pass


class NoiseRemovalWidget(qt.QWidget):
    """
    Widget to apply noise removal from a dataset.

    Background computation is done in a dedicated thread.

    All operations are done "on the fly" in the main thread on the current active image of the stack.

    For now it can apply both background subtraction and hot pixel removal and mask.
    For background subtraction the user can choose the background to use:
    dark frames, low intensity data or all the data. From these background
    frames, an image is computed either using the mean or the median.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self._incomingOperations: list[NoiseRemovalOperation] = []
        self._dataset: dtypes.Dataset | None = None
        self.setWindowFlags(qt.Qt.WindowType.Widget)

        self._parametersWidget = ParametersWidget()
        self._plot = Plot2D(self)

        self._browser = HorizontalSliderWithBrowser(self)
        self._browser.setRange(0, 0)
        self._browser.setEnabled(False)

        self._plot.setKeepDataAspectRatio(True)

        self._operationList = OperationListWidget()

        layout = qt.QVBoxLayout()
        layout.addWidget(self._plot)
        layout.addWidget(self._browser)
        layout.addWidget(self._parametersWidget)
        layout.addWidget(self._operationList)
        self.setLayout(layout)

        self._backgroundComputationThread: OperationThread | None = None
        self._previewThread = PreviewThread(self)

        # Add connections
        self._parametersWidget.computeBS.clicked.connect(
            self._launchBackgroundSubtraction
        )
        self._parametersWidget.computeHP.clicked.connect(self._launchHotPixelRemoval)
        self._parametersWidget.computeTP.clicked.connect(self._launchThresholdRemoval)
        self._parametersWidget.computeMR.clicked.connect(self._launchMaskRemoval)
        self._browser.valueChanged.connect(self.refreshPlot)
        self._operationList.sigOneOperationRemoved.connect(self.refreshPlot)
        self._previewThread.sigDataReady.connect(self._onOneFramePreviewReady)
        self._previewThread.finished.connect(self._onPreviewFinished)

    def setAppliedState(self, outputDataset: dtypes.Dataset):
        """
        Put widget on a state where every controls and operation list are disabled.
        Update dataset to `outputDataset`
        Operations are not applied online because we considere the input dataset was updated

        Call `setDataset` to exit from applied mode.
        """
        self._dataset = outputDataset
        self._operationList.setDisabled(True)
        self._parametersWidget.setDisabled(True)

    def setDataset(self, dataset: dtypes.Dataset):
        """Saves the dataset and updates the stack with the dataset data."""

        self._operationList.clear()
        self._plot.clear()

        self._dataset = dataset
        if self._imgDataset.title != "":
            self._plot.setTitleCallback(lambda idx: self._imgDataset.title)

        self._initPlot()

        self._parametersWidget.computeBS.show()
        self._parametersWidget.computeHP.show()
        self._parametersWidget.computeTP.show()
        self._parametersWidget.computeMR.show()

        self._parametersWidget.setEnabled(True)
        self._operationList.setEnabled(True)

        """
        Sets the available background for the user to choose.
        """
        self._parametersWidget.bsBackgroundCB.clear()
        if dataset.bg_dataset is not None:
            self._parametersWidget.bsBackgroundCB.addItem(
                BackgroundType.DARK_DATA.value
            )
        if dataset.bg_indices is not None:
            self._parametersWidget.bsBackgroundCB.addItem(
                BackgroundType.UNUSED_DATA.value
            )
        self._parametersWidget.bsBackgroundCB.addItem(BackgroundType.DATA.value)

    def _launchBackgroundSubtraction(self):
        self._computeBackgroundAsync(
            self._parametersWidget.bsBackgroundCB.currentText(),
            self._parametersWidget.bsMethodsCB.currentText(),
        )

    def _computeBackgroundAsync(self, background_type: str, method: str):

        assert (
            self._backgroundComputationThread is None
        ), "Computation thread is expected to be None."

        self._parametersWidget.setDisabled(True)
        self._parametersWidget.computeBS.setWaiting(True)

        # Here the background image is computed.
        self._backgroundComputationThread = OperationThread(
            self, create_background_substraction_operation
        )
        self._backgroundComputationThread.setArgs(
            self._dataset,
            method,
            background_type,
        )
        self._backgroundComputationThread.finished.connect(
            self._onBackgroundComputationFinish
        )
        self._backgroundComputationThread.start()

    def _onBackgroundComputationFinish(self):
        self._parametersWidget.setEnabled(True)
        self._parametersWidget.computeBS.setWaiting(False)

        if self._backgroundComputationThread.data is None:
            raise RuntimeError("An exception occured during background computation.")

        # Result of `create_background_substraction_operation` is a new operation that contains the pre-computed 2D background image
        operation = self._backgroundComputationThread.data

        self._backgroundComputationThread = None

        if operation is not None:
            self._operationList.append(operation)
            self.refreshPlot()

    def _launchHotPixelRemoval(self):
        size = self._parametersWidget.hpSizeCB.currentText()

        operation = NoiseRemovalOperation(
            type=Operation.HP,
            parameters={
                "kernel_size": int(size),
            },
        )
        self._operationList.append(operation)
        self.refreshPlot()

    def _launchThresholdRemoval(self):
        bottom_threshold = self._parametersWidget.bottomLE.text()
        top_threshold = self._parametersWidget.topLE.text()
        try:
            bottom_threshold = int(bottom_threshold)
        except ValueError:
            bottom_threshold = None
        try:
            top_threshold = int(top_threshold)
        except ValueError:
            top_threshold = None

        operation = NoiseRemovalOperation(
            type=Operation.THRESHOLD,
            parameters={
                "bottom": bottom_threshold,
                "top": top_threshold,
            },
        )
        self._operationList.append(operation)
        self.refreshPlot()

    def _launchMaskRemoval(self):
        mask = self._plotWidget.getSelectionMask()
        if mask is None:
            return
        operation = NoiseRemovalOperation(
            type=Operation.MASK,
            parameters={"mask": mask},
        )
        self._operationList.append(operation)
        self.refreshPlot()

    def refreshPlot(self):
        frameNumber = self._browser.value()
        img = self._plot.getActiveImage()
        if img is None:
            # No active image -> nothing to refresh
            return

        if (
            self._operationList.isEnabled()
            and len(self._operationList.getOperations()) > 0
        ):
            # processing on the fly
            data = self._imgDataset.as_array3d()[frameNumber].copy()
            self._previewThread.queue.put((data, self._operationList.getOperations()))
            if not self._previewThread.isRunning():
                # Re-start work if timeout reached
                self._previewThread.start()
        else:
            # show unprocessed images
            img.setData(self._imgDataset.as_array3d()[frameNumber], copy=False)

    def _onPreviewFinished(self):
        self._plot.clearMarkers()

    def _onOneFramePreviewReady(self, data):
        img = self._plot.getActiveImage()
        if img is None:
            return
        self._plot.clearMarkers()
        marker = self._plot.addMarker(
            self._plot.getGraphXLimits()[0],
            self._plot.getGraphYLimits()[1],
            symbol="",
            text="processing...",
            color="white",
        )
        marker.setBackgroundColor([0.0, 0.0, 0.0, 0.5])
        img.setData(data, copy=False)

    def _initPlot(self):
        if self._imgDataset is None:
            self._plot.clear()
            self._browser.setDisabled(True)
            return
        self._browser.setRange(0, self._imgDataset.nframes - 1)
        self._browser.setEnabled(True)
        self._plot.addImage(
            self._imgDataset.as_array3d()[0],
            colormap=Colormap(
                name=config.DEFAULT_COLORMAP_NAME,
                normalization=config.DEFAULT_COLORMAP_NORM,
            ),
        )

    def setDefaultParameters(self, operations: list[NoiseRemovalOperation]):
        """
        Set default values un widget parameters based on operations list

        """
        self._parametersWidget.set_default_values(operations)

    def getOperationList(self):
        return self._operationList.getOperations()

    def setOperationList(self, operations: list[NoiseRemovalOperation]):
        self._operationList.clear()
        self._operationList.extend(operations)

    def hasIncomingOperations(self) -> bool:
        return len(self._incomingOperations) > 0

    def getIncomingOperations(self) -> tuple[NoiseRemovalOperation]:
        return self._incomingOperations

    @property
    def _imgDataset(self) -> dtypes.ImageDataset:
        return self._dataset.dataset

    @property
    def _plotWidget(self):
        return self._plot



from __future__ import annotations

from silx.gui import qt

from ...core.noiseremoval import NoiseRemovalOperation
from ...core.noiseremoval import operation_to_str


class OperationListWidget(qt.QWidget):
    """Keeps the history of noise removal operations and displays them in a QListWidget"""

    sigOneOperationRemoved = qt.Signal()

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent)

        layout = qt.QVBoxLayout()
        assert layout is not None
        self._listWidget = qt.QListWidget()
        self._listWidget.setSelectionMode(
            qt.QAbstractItemView.SelectionMode.SingleSelection
        )
        self._listWidget.hide()
        self._checkbox = qt.QCheckBox("Show operations")

        layout.addWidget(self._checkbox)
        layout.addWidget(self._listWidget)
        self.setLayout(layout)

        self._checkbox.toggled.connect(self._listWidget.setVisible)
        self._checkbox.setChecked(True)

        self._operations: list[NoiseRemovalOperation] = []

        self._listWidget.setContextMenuPolicy(
            qt.Qt.ContextMenuPolicy.ActionsContextMenu
        )

        deleteAction = qt.QAction("Delete", self)
        deleteAction.setShortcut(qt.QKeySequence.StandardKey.Delete)
        self._listWidget.addAction(deleteAction)

        # Connect Signal / Slot
        deleteAction.triggered.connect(self.__removeSelection)

    def append(self, operation: NoiseRemovalOperation):
        self._operations.append(operation)
        self._listWidget.addItem(operation_to_str(operation))

    def clear(self):
        self._operations.clear()
        self._listWidget.clear()

    def getOperations(self) -> list[NoiseRemovalOperation]:
        return list(self._operations)

    def extend(self, operations: list[NoiseRemovalOperation]):
        self._operations.extend(operations)
        for i_new_op, operation in enumerate(operations):
            self._listWidget.addItem(qt.QListWidgetItem(operation_to_str(operation)))

    def __removeSelection(self):
        """Remove selected operation (triggered by button or Delete key)."""
        row = self._listWidget.currentRow()
        if row < 0:
            return

        # remove from data
        del self._operations[row]
        # remove from list widget
        self._listWidget.removeItemWidget(self._listWidget.takeItem(row))

        self.sigOneOperationRemoved.emit()



import tempfile

import h5py
import numpy
from silx.io.dictdump import dicttonx
from silx.io.nxdata import NXdata

from darfix.core.grainplot import MultiDimMomentType
from darfix.core.grainplot import OrientationDistData
from darfix.core.grainplot import generate_grain_maps_nxdict
from darfix.core.moment_types import MomentType

from .utils import createHDF5Dataset1D
from .utils import createHDF5Dataset3D


def test_generate_grain_maps_nxdict_1d():

    dset = createHDF5Dataset1D(numpy.random.rand(10, 10, 10))
    dset.find_dimensions()
    dset.apply_moments()
    nxdict = generate_grain_maps_nxdict(dset, None, None)

    for moment_type in MomentType:
        assert moment_type.value in nxdict["entry"]
        assert nxdict["entry"][moment_type.value][moment_type.value].shape == (10, 10)

    assert MultiDimMomentType.ORIENTATION_DIST.value not in nxdict["entry"]
    assert MultiDimMomentType.MOSAICITY.value not in nxdict["entry"]


def test_generate_grain_maps_nxdict_3d():
    dset = createHDF5Dataset3D(numpy.random.rand(5, 5, 5, 6, 6))
    dset.find_dimensions()
    dset.apply_moments()
    orientation_dist_data = OrientationDistData(dset, 0, 1, [0, 1], [0, 1], dset.zsum())
    mosa = numpy.zeros((10, 10))

    nxdict = generate_grain_maps_nxdict(dset, mosa, orientation_dist_data)

    for motor in ("motor1", "motor2", "motor3"):
        for moment_type in MomentType:
            assert moment_type.value in nxdict["entry"][motor]
            assert nxdict["entry"][motor][moment_type.value][
                moment_type.value
            ].shape == (6, 6)

    assert MultiDimMomentType.ORIENTATION_DIST.value in nxdict["entry"]
    assert MultiDimMomentType.MOSAICITY.value in nxdict["entry"]

    with tempfile.TemporaryFile() as f:
        dicttonx(nxdict, f)

        with h5py.File(f) as h5f:
            assert NXdata(
                h5f["entry"][MultiDimMomentType.ORIENTATION_DIST.value]["data"]
            ).is_image
            assert NXdata(
                h5f["entry"][MultiDimMomentType.ORIENTATION_DIST.value]["key"]
            ).is_image
            assert NXdata(h5f["entry"][MultiDimMomentType.MOSAICITY.value]).is_image



from __future__ import annotations

from darfix import dtypes
from darfix.core.grainplot import GrainPlotMaps
from darfix.gui.grainplot.grainPlotWidget import GrainPlotWidget
from darfix.tasks.grainplot import GrainPlot
from orangecontrib.darfix.widgets.datasetWidgetBase import DatasetWidgetBase


class GrainPlotWidgetOW(DatasetWidgetBase, ewokstaskclass=GrainPlot):
    """
    Computes moments (Center of mass, FWHM, Kurtosis, SKEWNESS) and displays them.

    Also computes mosaicity and orientation distribution for multi-dimensional datasets.
    """

    _ewoks_inputs_to_hide_from_orange = (
        "filename",
        "dimensions",
        "range",
        "save_maps",
        "third_motor",
        "orientation_img_origin",
    )

    name = "grain plot"
    icon = "icons/grainplot.png"
    description = "Computes Center of mass, FWHM, Kurtosis, Skewness, mosaicity (nD), orientation (nD) maps and displays them"
    want_main_area = True
    want_control_area = False

    def __init__(self):
        super().__init__()

        self._widget = GrainPlotWidget(parent=self)
        self.mainArea.layout().addWidget(self._widget)
        self.datasetChanged.connect(self.onDatasetChanged)

    def get_task_inputs(self):
        task_inputs = super().get_task_inputs()

        # Saving is handled by the widget
        task_inputs["save_maps"] = False

        return task_inputs

    def onDatasetChanged(self, dataset: dtypes.Dataset):

        self._widget.setMessage("Computing...")
        try:
            self.execute_ewoks_task()
        except Exception as e:
            self._widget.setMessage(f"Error while computing: {e}!")
            raise e

    def task_output_changed(self):
        dataset = self.get_task_output_value("dataset", None)

        if dataset is None:
            return

        if not isinstance(dataset, dtypes.Dataset):
            raise dtypes.DatasetTypeError(dataset)
        self._widget.setMessage("Computing finished!")
        self._widget.setGrainPlotMaps(GrainPlotMaps.from_dataset(dataset))



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from enum import Enum as _Enum
from typing import Any
from typing import Callable
from typing import TypedDict

import numpy
from tqdm import tqdm

from darfix.core.utils import OperationAborted

from ..dtypes import Dataset
from ..processing import imageOperations
from .state_of_operation import Operation


class BackgroundType(_Enum):
    DATA = "Data"
    UNUSED_DATA = "Unused data (after partition)"
    DARK_DATA = "Dark data"


class NoiseRemovalOperation(TypedDict):
    type: Operation
    parameters: dict[str, Any]


def operation_to_str(op: NoiseRemovalOperation) -> str:
    result = None
    op_type = op["type"]
    if op_type is Operation.BS:
        result = f"Background subtraction {op['parameters']}"

    elif op_type is Operation.HP:
        result = f"Hot pixel removal: {op['parameters']}"

    elif op_type is Operation.THRESHOLD:
        result = f"Threshold removal: {op['parameters']}"

    elif op_type is Operation.MASK:
        result = "Mask removal"
    else:
        raise ValueError(f"Unknown operation type {op_type}")
    return result


def clean_operation_dict(op: NoiseRemovalOperation) -> dict:
    """Just keep `type` and `parameters` in the saved dict"""
    return {
        "type": op["type"],
        "parameters": op["parameters"],
    }


def apply_noise_removal_operation(
    image: numpy.ndarray, operation: NoiseRemovalOperation
) -> None:
    if operation["type"] is Operation.BS:
        apply_background_substraction(image, operation["background"])

    if operation["type"] is Operation.HP:
        apply_hot_pixel_removal(image, **operation["parameters"])

    if operation["type"] is Operation.THRESHOLD:
        apply_threshold_removal(image, **operation["parameters"])

    if operation["type"] is Operation.MASK:
        apply_mask_removal(image, **operation["parameters"])


def create_background_substraction_operation(
    dataset: Dataset,
    method: str | None = None,
    background_type: BackgroundType | None = None,
) -> NoiseRemovalOperation:
    """Compute a background image and return a NoiseRemovalOperation with a background parameter"""

    if method is None:
        method = imageOperations.Method.MEDIAN

    if background_type is not None:
        background_type = BackgroundType(background_type)

    if background_type == BackgroundType.DARK_DATA:
        bg = dataset.bg_dataset.as_array3d()
    elif background_type == BackgroundType.UNUSED_DATA:
        bg = dataset.dataset.as_array3d(dataset.bg_indices)
    elif background_type == BackgroundType.DATA:
        bg = dataset.dataset.as_array3d()
    else:
        raise NotImplementedError(
            f"Background type {background_type!r} not implemented yet."
        )
    parameters = dict(method=method, background_type=background_type)

    return NoiseRemovalOperation(
        type=Operation.BS,
        parameters=parameters,
        background=imageOperations.compute_background(bg, method),
    )


def apply_background_substraction(
    image: numpy.ndarray, background: numpy.ndarray
) -> None:
    imageOperations.background_subtraction(image, background)


def apply_hot_pixel_removal(
    image: numpy.ndarray, kernel_size: int | None = None
) -> None:
    if kernel_size is None:
        kernel_size = 3

    imageOperations.hot_pixel_removal(image, ksize=kernel_size)


def apply_threshold_removal(
    image: numpy.ndarray, bottom: int | None = None, top: int | None = None
) -> None:
    imageOperations.threshold_removal(image, bottom=bottom, top=top)


def apply_mask_removal(image: numpy.ndarray, mask: numpy.ndarray | None) -> None:
    if mask is None:
        return
    imageOperations.mask_removal(image, mask)


def apply_noise_removal_operations(
    image_stack: numpy.ndarray,
    operations: list[NoiseRemovalOperation],
    is_cancel: Callable[[], bool],
    set_progress: Callable[[int], None],
    max_workers: int | None = None,
) -> None:
    if image_stack.ndim != 3:
        raise ValueError(
            f"image_stack must have ndim == 3 but ndims is {image_stack.ndim}"
        )

    def process_single(img: numpy.ndarray):
        if is_cancel():
            return
        for op in operations:
            try:
                apply_noise_removal_operation(img, operation=op)
            except Exception as e:
                # Return e to re-raise Exception in the parent thread
                return e

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = [executor.submit(process_single, img) for img in image_stack]
        len_futures = len(futures)
        for idx_future, future in tqdm(
            enumerate(as_completed(futures)),
            desc="Noise removal operation",
            total=len_futures,
        ):
            exception = future.result()
            if isinstance(exception, Exception):
                raise exception
            if is_cancel():
                raise OperationAborted()
            set_progress(idx_future * 100 // len_futures)



from __future__ import annotations

import weakref

import numpy
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot import PlotWidget
from silx.gui.plot.actions.control import KeepAspectRatioAction
from silx.gui.plot.items.image import ImageBase
from silx.gui.plot.items.marker import Marker
from silx.gui.plot.items.roi import RectangleROI
from silx.gui.plot.tools import InteractiveModeToolBar
from silx.gui.plot.tools import OutputToolBar
from silx.gui.plot.tools.roi import RegionOfInterestManager

from .flashlight_mode_action import FlashlightModeAction


def _toImgCoord(img: ImageBase, coord: tuple[float, float]) -> tuple[int, int] | None:
    if img is None:
        # Not init
        return None
    origin = numpy.asarray(img.getOrigin())
    scale = numpy.asarray(img.getScale())

    i, j = tuple(((numpy.asarray(coord) - origin) / scale).astype(int))

    if 0 > i or i >= img.getData(False).shape[1]:
        # out of bound
        return None
    if 0 > j or j >= img.getData(False).shape[0]:
        # out of bound
        return None
    return i, j


class RectangleImageROI(RectangleROI):
    def __init__(self, plot: PlotWidget):
        super().__init__()
        self.__plotRef = weakref.ref(plot)
        self.__centerPositionLabel: Marker = self.addLabelHandle()
        self.__centerPositionLabel.setBackgroundColor([1.0, 1.0, 1.0, 0.5])
        self.sigRegionChanged.connect(self._onRegionChanged)

    @property
    def plot(self) -> PlotWidget:
        """The :class:`PlotWidget` this roi is attached."""
        plot = self.__plotRef()
        if plot is None:
            raise ValueError("The attached plot was already destroyed")
        return plot

    def _onRegionChanged(self):

        x, y = self.getOrigin()
        self.__centerPositionLabel.setPosition(x, y)
        x, y = self.getCenter()
        sizeX, sizeY = self.getSize()
        self.__centerPositionLabel.setText(
            f"Center: {x:.3f}, {y:.3f} \n(Size : {sizeX:.3f}, {sizeY:.3f})"
        )

    def getCenterIJ(self) -> tuple[int, int]:
        """ROI center as indices i,j in the underlying image data"""
        coordinates_ij = _toImgCoord(self.plot.getActiveImage(), self.getCenter())
        if coordinates_ij is None:
            raise ValueError("ROI center outside of the plot or no active image")
        return coordinates_ij

    def getSlices(self) -> tuple:
        """ROI slices give the image data array selection as a tuple of two slices over i and j axis"""

        if self.plot.getImage() is None:
            raise ValueError(
                "At least one image should be added in InteractiveMaskPlotWithFlashlight"
            )
        scale = self.plot.getActiveImage().getScale()

        coordinates_ij = _toImgCoord(self.plot.getActiveImage(), self.getCenter())
        if coordinates_ij is not None:
            i, j = coordinates_ij
        else:
            raise ValueError("ROI center outside of the plot or no active image")

        w = int(self.getSize()[0] / scale[0])
        h = int(self.getSize()[1] / scale[1])

        return (slice(max(0, j - h), j + h), slice(max(0, i - w), i + w))


class PlotWithFlashlight(PlotWidget):
    """
    A plot with a roi that follow mouse. Shape of the ROI can be modified with button left pressed.
    """

    DEFAULT_FLASHLIGHT_SIZE = 0.1

    sigFlashlightHide = qt.Signal()
    sigFlashlightMove = qt.Signal()

    def __init__(self, **kwarks):
        super().__init__(**kwarks)
        self.__selectionStart = None

        self.__roiManager = RegionOfInterestManager(parent=self)
        self.__roi = RectangleImageROI(self)

        self.__roiManager.addRoi(self.__roi)
        self.__roi.setSize((self.DEFAULT_FLASHLIGHT_SIZE, self.DEFAULT_FLASHLIGHT_SIZE))
        self.__roi.setVisible(False)
        self.__roi.setColor([0.2, 0.2, 0.2])
        self.__roi.setLineStyle("--")

        self.__interactiveToolBar = InteractiveModeToolBar(plot=self)
        self.__interactiveToolBar.addAction(
            FlashlightModeAction(parent=self, plot=self)
        )

        outputToolBar = OutputToolBar(plot=self)
        outputToolBar.addAction(KeepAspectRatioAction(parent=self, plot=self))

        self.addToolBar(self.__interactiveToolBar)
        self.addToolBar(outputToolBar)

        self.setInteractiveMode(FlashlightModeAction.INTERACTION_MODE)

        self.setMouseTracking(True)

        self.sigPlotSignal.connect(self._onPlotSignal)

    def getRoi(self) -> RectangleImageROI:
        """Get the current ROI"""
        return self.__roi

    def isFlashlightEnabled(self) -> bool:
        """Is flashlight mode enabled"""
        return (
            self.getInteractiveMode()["mode"] == FlashlightModeAction.INTERACTION_MODE
        )

    def _onPlotSignal(self, evt: dict):
        if evt["event"] != "mouseMoved":
            return
        dataCoordinates = (evt["x"], evt["y"])
        imgCoordinates = _toImgCoord(self.getActiveImage(), dataCoordinates)
        if imgCoordinates is not None:
            self.__onMove(*imgCoordinates, *dataCoordinates)
        else:
            self._leave()

    def _leave(self):
        if self.__roi.isVisible():
            self.sigFlashlightHide.emit()
            self.setCursor(qt.Qt.CursorShape.ArrowCursor)
            self.__roi.setVisible(False)

    def __onMove(self, i: int, j: int, x: float, y: float):
        if not self.isFlashlightEnabled():
            return
        if not self.__roi.isVisible():
            self.__roi.setVisible(True)
        if self.__selectionStart is not None:
            newRoiSize = numpy.abs(numpy.subtract((x, y), self.__selectionStart)) * 2
            self.__roi.setSize(newRoiSize)
        self.setCursor(qt.Qt.CursorShape.CrossCursor)
        self.__roi.setCenter((x, y))
        self.sigFlashlightMove.emit()

    def onMousePress(self, xPixel, yPixel, btn):
        super().onMousePress(xPixel, yPixel, btn)
        if not self.isFlashlightEnabled():
            return
        dataCoordinates = self.pixelToData(xPixel, yPixel, check=True)
        if dataCoordinates is not None:
            self.__selectionStart = dataCoordinates

    def onMouseRelease(self, xPixel, yPixel, btn):
        self.__selectionStart = None
        super().onMouseRelease(xPixel, yPixel, btn)

    def onMouseLeaveWidget(self):
        self._leave()
        super().onMouseLeaveWidget()

    def _getMaskImage(self, legend: str, alpha_value=127):
        img = self.getImage(legend)
        if img is None:
            active = self.getActiveImage()
            data = numpy.zeros(self.getPlotMaskShape(), numpy.uint8)
            img = self.addImage(
                data,
                legend,
                scale=active.getScale(),
                origin=active.getOrigin(),
                colormap=Colormap(colors=[[0, 0, 0, 0], [0, 0, 0, alpha_value]]),
            )
        return img

    def clearPlotMask(self, legend: str = "__MASK__"):
        """in the plot, clear image mask with legend `legend`"""
        if self.getImage(legend) is not None:
            self.removeImage(legend)

    def createPlotMask(self, mask: numpy.ndarray, legend: str = "__MASK__"):
        """in the plot, create image mask with legend `legend`"""
        self._getMaskImage(legend).setData(mask, copy=False)

    def getPlotMaskShape(self) -> tuple[int, int]:
        """Get desired 2D shape to fit shape of active image in the plot"""
        return self.getActiveImage().getData(False).shape[:2]



from __future__ import annotations

import logging

from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot import Plot2D
from silx.io.dictdump import dicttonx

from darfix.core.moment_types import MomentType

from ...core.grainplot import GrainPlotMaps
from ...core.grainplot import generate_grain_maps_nxdict
from ..utils.utils import select_output_hdf5_file_with_dialog
from .mosaicityWidget import MosaicityWidget
from .utils import MapType
from .utils import add_image_with_transformation

_logger = logging.getLogger(__file__)


class GrainPlotWidget(qt.QMainWindow):
    """
    Widget to show a series of maps for the analysis of the data.
    """

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        self._grainPlotMaps: GrainPlotMaps | None = None
        self._plots: list[Plot2D] = []

        self._mapTypeComboBox = qt.QComboBox()
        self._mapTypeComboBox.addItems([map_type.value for map_type in MapType])
        self._mapTypeComboBox.currentTextChanged.connect(self._updatePlot)

        self._plotWidget = qt.QWidget()
        plotsLayout = qt.QHBoxLayout()
        self._plotWidget.setLayout(plotsLayout)
        widget = qt.QWidget(parent=self)
        layout = qt.QVBoxLayout()

        self._mosaicity_widget = MosaicityWidget()

        self._exportButton = qt.QPushButton("Export maps")
        self._exportButton.setDefault(False)
        self._exportButton.clicked.connect(self.exportMaps)

        self._messageWidget = qt.QLabel("No dataset in input.")

        layout.addWidget(self._mapTypeComboBox)
        layout.addWidget(self._plotWidget)
        layout.addWidget(self._mosaicity_widget)
        layout.addWidget(self._messageWidget)
        layout.addWidget(self._exportButton)
        self._plotWidget.hide()
        self._mosaicity_widget.hide()
        widget.setLayout(layout)
        self.setCentralWidget(widget)

        self._mapTypeComboBox.setDisabled(True)
        self._exportButton.setDisabled(True)

    def setMessage(self, message: str):
        self._messageWidget.setText(message)

    def setGrainPlotMaps(self, grainPlotMaps: GrainPlotMaps):

        self._grainPlotMaps = grainPlotMaps

        for i in reversed(range(self._plotWidget.layout().count())):
            self._plotWidget.layout().itemAt(i).widget().setParent(None)

        self._plots.clear()
        for dim in self._grainPlotMaps.dims.values():
            plot = Plot2D(parent=self)
            plot.setKeepDataAspectRatio(True)
            plot.setGraphTitle(self._grainPlotMaps.title + "\n" + dim.name)
            plot.setDefaultColormap(Colormap(name="viridis"))
            self._plots.append(plot)
            self._plotWidget.layout().addWidget(plot)

        # Enable mosaicity if ndim >= 2
        if self._grainPlotMaps.dims.ndim >= 2:
            self._mosaicity_widget.setGrainPlotMaps(self._grainPlotMaps)
            self._mapTypeComboBox.model().item(4).setEnabled(True)
            self._mapTypeComboBox.setCurrentText(MapType.MOSAICITY.value)
        else:
            self._mapTypeComboBox.model().item(4).setEnabled(False)
            self._mapTypeComboBox.setCurrentText(MapType.COM.value)
        # Force plot update since setCurrentText does not fire currentTextChanged if the text is the same
        # https://doc.qt.io/qt-6/qcombobox.html#currentTextChanged
        self._updatePlot(self._mapTypeComboBox.currentText())

        self._messageWidget.hide()
        self._mapTypeComboBox.setEnabled(True)
        self._exportButton.setEnabled(True)
        self._exportButton.setDefault(False)

    def _updatePlot(self, raw_map_type: str):
        """
        Update shown plots in the widget
        """

        map_type = MapType(raw_map_type)
        if map_type == MapType.MOSAICITY:
            self._mosaicity_widget.show()
            self._plotWidget.hide()
            return

        if self._grainPlotMaps is None:
            return

        moment_type = MomentType(raw_map_type)

        moments = self._grainPlotMaps.moments_dims
        self._mosaicity_widget.hide()
        self._plotWidget.show()
        for i, plot in enumerate(self._plots):
            self._addImage(plot, moments[i][moment_type])

    def _generate_maps_nxdict(self) -> dict:
        orientation_dist_data = self._mosaicity_widget.getOrientationDist()

        return generate_grain_maps_nxdict(
            self._grainPlotMaps,
            self._mosaicity_widget.getMosaicity(),
            orientation_dist_data,
        )

    def exportMaps(self):
        """
        Creates dictionary with maps information and exports it to a nexus file
        """
        nx = self._generate_maps_nxdict()

        filename = select_output_hdf5_file_with_dialog()
        if filename:
            dicttonx(nx, filename)

    def _addImage(self, plot, image):
        if self._grainPlotMaps is None:
            transformation = None
        else:
            transformation = self._grainPlotMaps.transformation
        add_image_with_transformation(plot, image, transformation)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any
from typing import Dict
from typing import Tuple

import numpy
from matplotlib.colors import hsv_to_rgb
from silx.math.medianfilter import medfilt2d

from darfix.core.dimension import AcquisitionDims
from darfix.core.transformation import Transformation
from darfix.io.utils import create_nxdata_dict

from ..dtypes import AxisType
from ..dtypes import Dataset
from .moment_types import MomentType
from .utils import compute_hsv

DimensionRange = Tuple[float, float]


DimensionMoments = Dict[MomentType, numpy.ndarray]


@dataclass
class GrainPlotMaps:
    dims: AcquisitionDims
    moments_dims: list[DimensionMoments]
    zsum: numpy.ndarray
    transformation: numpy.ndarray
    title: str

    @staticmethod
    def from_dataset(dataset: Dataset) -> GrainPlotMaps:
        """
        Just retrieve all necessary attributes in `ImageDataset` object to return a `GrainPlotMaps` object.

        :Note:  zsum computation is executed
        """
        imgDataset = dataset.dataset
        return GrainPlotMaps(
            dims=imgDataset.dims,
            moments_dims=imgDataset.moments_dims,
            zsum=imgDataset.zsum(),
            transformation=imgDataset.transformation,
            title=imgDataset.title,
        )


class OrientationDistData:
    KEY_IMAGE_SIZE = 1000

    def __init__(
        self,
        grain_plot_maps: GrainPlotMaps,
        x_dimension: int,
        y_dimension: int,
        x_dimension_range: DimensionRange,
        y_dimension_range: DimensionRange,
        zsum: numpy.ndarray | None = None,
    ) -> None:
        """
        Store data for orientation distribution RGB layer (the HSV colormap) and data (Histogram 2D)

        :param zsum: Precomputed dataset.zsum() used as the weight of the histogram 2D . If None dataset.zsum() is called.
        """

        if len(grain_plot_maps.moments_dims) == 0:
            raise ValueError("Moments should be computed before.")

        com_x = grain_plot_maps.moments_dims[x_dimension][MomentType.COM].ravel()
        com_y = grain_plot_maps.moments_dims[y_dimension][MomentType.COM].ravel()

        self.x_range = x_dimension_range
        self.y_range = y_dimension_range

        if zsum is None:
            zsum = grain_plot_maps.zsum
            zsum = zsum.ravel()
        else:
            zsum = zsum.ravel()

        # automatic bins
        # In darfix<=3.x orientation distribution shape was the size of the dimension
        # A x2 Factor is a little thinner. To see if it needs to be update in the future.
        self.x_bins = grain_plot_maps.dims.get(x_dimension).size * 2
        self.y_bins = grain_plot_maps.dims.get(y_dimension).size * 2

        self.x_label = grain_plot_maps.dims.get(x_dimension).name
        self.y_label = grain_plot_maps.dims.get(y_dimension).name

        x_data = numpy.linspace(-1, 1, self.KEY_IMAGE_SIZE)
        y_data = numpy.linspace(-1, 1, self.KEY_IMAGE_SIZE)
        x_mesh, y_mesh = numpy.meshgrid(x_data, y_data)
        self.rgb_key = hsv_to_rgb(compute_hsv(x_mesh, y_mesh))

        # Histogram in 2D
        histogram, _, _ = numpy.histogram2d(
            com_y,
            com_x,  # note: y first is in purpose : see numpy.histogram2d documentation
            weights=zsum,  # We need to take into account pixel intensity
            bins=[self.y_bins, self.x_bins],
            range=[self.y_range, self.x_range],
        )
        self.data = histogram
        """Orientation distribution data as an histogram 2D of the center of mass in two dimensions"""
        self.smooth_data = medfilt2d(numpy.ascontiguousarray(self.data))
        """ `self.data` filtered with a median filter 2D"""

    def x_data_values(self) -> numpy.ndarray:
        return numpy.linspace(
            self.x_range[0], self.x_range[1], self.x_bins, endpoint=False
        )

    def y_data_values(self) -> numpy.ndarray:
        return numpy.linspace(
            self.y_range[0], self.y_range[1], self.y_bins, endpoint=False
        )

    def x_rgb_key_values(self) -> numpy.ndarray:
        return numpy.linspace(
            self.x_range[0], self.x_range[1], self.KEY_IMAGE_SIZE, endpoint=False
        )

    def y_rgb_key_values(self) -> numpy.ndarray:
        return numpy.linspace(
            self.y_range[0], self.y_range[1], self.KEY_IMAGE_SIZE, endpoint=False
        )

    def origin(
        self,
        origin: AxisType,
    ) -> tuple[float, float]:
        if origin == "dims":
            return (self.x_range[0], self.y_range[0])
        elif origin == "center":
            return (
                -numpy.ptp(self.x_range) / 2,
                -numpy.ptp(self.y_range) / 2,
            )
        else:
            return (0, 0)

    def data_plot_scale(self) -> tuple[float, float]:
        return (
            numpy.ptp(self.x_range) / self.x_bins,
            numpy.ptp(self.y_range) / self.y_bins,
        )

    def rgb_key_plot_scale(self) -> tuple[float, float]:
        return (
            numpy.ptp(self.x_range) / self.KEY_IMAGE_SIZE,
            numpy.ptp(self.y_range) / self.KEY_IMAGE_SIZE,
        )

    def to_motor_coordinates(
        self,
        points_x: numpy.ndarray,
        points_y: numpy.ndarray,
        origin: AxisType,
    ) -> tuple[numpy.ndarray, numpy.ndarray]:
        """
        Given points_x, points_y in the 2D space of self.data, returns motor coordinates x, y
        """
        x_origin, y_origin = self.origin(origin)
        return (
            points_x * numpy.ptp(self.x_range) / (self.x_bins - 1) + x_origin,
            points_y * numpy.ptp(self.y_range) / (self.y_bins - 1) + y_origin,
        )


class MultiDimMomentType(Enum):
    """Moments that are only computed for datasets with multiple dimensions"""

    ORIENTATION_DIST = "Orientation distribution"
    MOSAICITY = "Mosaicity"


def get_axes(transformation: Transformation | None) -> tuple[
    tuple[numpy.ndarray, numpy.ndarray] | None,
    tuple[str, str] | None,
    tuple[str, str] | None,
]:
    if not transformation:
        return None, None, None

    axes = (transformation.xregular, transformation.yregular)
    axes_names = ("x", "y")
    axes_long_names = (transformation.label, transformation.label)

    return axes, axes_names, axes_long_names


def compute_normalized_component(
    component: numpy.ndarray, dimension_range: DimensionRange
):

    min_component = dimension_range[0]
    max_component = dimension_range[1]

    return 2 * (component - min_component) / (max_component - min_component) - 1


def compute_mosaicity(
    moments: dict[int, numpy.ndarray],
    x_dimension: int,
    y_dimension: int,
    x_dimension_range: DimensionRange,
    y_dimension_range: DimensionRange,
):
    norm_center_of_mass_x = compute_normalized_component(
        moments[x_dimension][MomentType.COM],
        dimension_range=x_dimension_range,
    )
    norm_center_of_mass_y = compute_normalized_component(
        moments[y_dimension][MomentType.COM],
        dimension_range=y_dimension_range,
    )
    return hsv_to_rgb(compute_hsv(norm_center_of_mass_x, norm_center_of_mass_y))


def create_moment_nxdata_groups(
    parent: dict[str, Any],
    moment_data: numpy.ndarray,
    axes,
    axes_names,
    axes_long_names,
):

    for map_type in MomentType:
        map_value = map_type.value
        parent[map_value] = create_nxdata_dict(
            moment_data[map_type],
            map_value,
            axes,
            axes_names,
            axes_long_names,
        )


def generate_grain_maps_nxdict(
    grainPlotMaps: GrainPlotMaps,
    mosaicity: numpy.ndarray | None,
    orientation_dist_image: OrientationDistData | None,
) -> dict:
    moments = grainPlotMaps.moments_dims
    axes, axes_names, axes_long_names = get_axes(grainPlotMaps.transformation)

    nx = {
        "entry": {"@NX_class": "NXentry"},
        "@NX_class": "NXroot",
        "@default": "entry",
    }

    if mosaicity is not None:
        nx["entry"][MultiDimMomentType.MOSAICITY.value] = create_nxdata_dict(
            mosaicity,
            MultiDimMomentType.MOSAICITY.value,
            axes,
            axes_names,
            axes_long_names,
            rgba=True,
        )
        nx["entry"]["@default"] = MultiDimMomentType.MOSAICITY.value
    else:
        nx["entry"]["@default"] = MomentType.COM.value

    if orientation_dist_image is not None:
        nx["entry"][MultiDimMomentType.ORIENTATION_DIST.value] = {
            "key": create_nxdata_dict(
                orientation_dist_image.rgb_key,
                "image",
                (
                    orientation_dist_image.y_rgb_key_values(),
                    orientation_dist_image.x_rgb_key_values(),
                ),
                (orientation_dist_image.y_label, orientation_dist_image.x_label),
                rgba=True,
            ),
            "data": create_nxdata_dict(
                orientation_dist_image.data,
                "orientation distribution",
                (
                    orientation_dist_image.y_data_values(),
                    orientation_dist_image.x_data_values(),
                ),
                (orientation_dist_image.y_label, orientation_dist_image.x_label),
            ),
            "@default": "data",
        }

    if grainPlotMaps.dims.ndim <= 1:
        create_moment_nxdata_groups(
            nx["entry"],
            moments[0],
            axes,
            axes_names,
            axes_long_names,
        )
    else:
        for axis, dim in grainPlotMaps.dims.items():
            nx["entry"][dim.name] = {"@NX_class": "NXcollection"}
            create_moment_nxdata_groups(
                nx["entry"][dim.name],
                moments[axis],
                axes,
                axes_names,
                axes_long_names,
            )

    return nx



from __future__ import annotations

from typing import List

import numpy
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot import Plot2D
from silx.io.dictdump import dicttonx

from darfix.gui.utils.custom_doublespinbox import createCustomDoubleSpinBox

from ..dtypes import Dataset
from ..io.utils import create_nxdata_dict
from ..pixel_sizes import PixelSize
from .utils.utils import select_output_hdf5_file_with_dialog


class _LineEditsWidget(qt.QWidget):
    def __init__(self, parent, dims=1, validator=None, placeholder=None):
        qt.QWidget.__init__(self, parent)
        layout = qt.QHBoxLayout()
        if placeholder:
            assert len(placeholder) == dims
        self._lineEdits = []
        for i in range(dims):
            lineEdit = qt.QLineEdit(parent=self)
            if placeholder:
                lineEdit.setPlaceholderText(placeholder[i])
            if validator:
                lineEdit.setValidator(validator)
            self._lineEdits.append(lineEdit)
            layout.addWidget(lineEdit)

        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    @property
    def value(self):
        values = []
        for le in self._lineEdits:
            values += [le.text() if le.text() != "" else None]
        return values

    def setValue(self, values: List[int]):
        assert type(values) is list

        for i, value in enumerate(values):
            self._lineEdits[i].setText(str(value))
            self._lineEdits[i].setCursorPosition(0)


class RSMHistogramWidget(qt.QWidget):
    """
    Widget to compute Reciprocal Space Map
    """

    sigComputeClicked = qt.Signal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.dataset: Dataset | None = None

        self._idx = [(2, 1), (2, 0), (1, 0)]

        layout = qt.QGridLayout()
        label = qt.QLabel("Q:")
        font = label.font()
        font.setBold(True)
        label.setFont(font)
        layout.addWidget(label, 0, 0, 1, 1)
        self._q = _LineEditsWidget(self, 3, qt.QIntValidator(), ["h", "k", "l"])
        self._q.setValue([1, 0, 1])
        layout.addWidget(self._q, 0, 1, 1, 2)
        label = qt.QLabel("Pseudocubic lattice:")
        label.setFont(font)
        layout.addWidget(label, 1, 0, 1, 2)
        self._a = createCustomDoubleSpinBox()
        self.a = 4.08
        layout.addWidget(self._a, 1, 1, 1, 2)
        label = qt.QLabel("Map range:")
        label.setFont(font)
        layout.addWidget(label, 2, 0, 1, 2)
        self._map_range = createCustomDoubleSpinBox()
        self.map_range = 0.008
        layout.addWidget(self._map_range, 2, 1, 1, 2)
        label = qt.QLabel("Detector:")
        label.setFont(font)
        layout.addWidget(label, 3, 0, 1, 2)
        self._detector = qt.QComboBox()
        self._detector.addItems(tuple(member.name for member in PixelSize))
        layout.addWidget(self._detector, 3, 1, 1, 2)
        label = qt.QLabel("Units:")
        label.setFont(font)
        layout.addWidget(label, 4, 0, 1, 2)
        self._units = qt.QComboBox()
        self._units.addItems(["Poulsen", "Gorfman"])
        layout.addWidget(self._units, 4, 1, 1, 2)
        label = qt.QLabel("n:")
        label.setFont(font)
        layout.addWidget(label, 5, 0, 1, 1)
        self._n = _LineEditsWidget(self, 3, qt.QIntValidator(), ["h", "k", "l"])
        self._n.setValue([0, 1, 0])
        layout.addWidget(self._n, 5, 1, 1, 2)
        label = qt.QLabel("Map shape:")
        label.setFont(font)
        layout.addWidget(label, 6, 0, 1, 1)
        self._map_shape = _LineEditsWidget(self, 3, qt.QIntValidator(), ["x", "y", "z"])
        self._map_shape.setValue([200, 200, 200])
        layout.addWidget(self._map_shape, 6, 1, 1, 2)
        label = qt.QLabel("Energy:")
        label.setFont(font)
        layout.addWidget(label, 7, 0, 1, 2)
        self._energy = createCustomDoubleSpinBox()
        self.energy = 17
        layout.addWidget(self._energy, 7, 1, 1, 2)
        self._computeB = qt.QPushButton("Compute")
        layout.addWidget(self._computeB, 8, 2, 1, 1)
        self._computeB.clicked.connect(self.sigComputeClicked.emit)
        self._plotWidget = qt.QWidget()
        self._plotWidget.setLayout(qt.QHBoxLayout())
        layout.addWidget(self._plotWidget, 9, 0, 1, 3)
        self._plotWidget.hide()
        self._exportButton = qt.QPushButton("Export maps")
        self._exportButton.hide()
        layout.addWidget(self._exportButton, 10, 2, 1, 1)
        self._exportButton.clicked.connect(self.exportMaps)

        self.setLayout(layout)
        self.setMinimumWidth(650)

    def setDataset(self, dataset: Dataset):
        self.dataset = dataset
        for i in reversed(range(self._plotWidget.layout().count())):
            self._plotWidget.layout().itemAt(i).widget().setParent(None)

        self._plotWidget.hide()
        self._exportButton.hide()

        self._plots = []
        for i in range(len(self._idx)):
            self._plots += [Plot2D(parent=self)]
            self._plots[-1].setDefaultColormap(Colormap(name="viridis"))
            self._plotWidget.layout().addWidget(self._plots[-1])

    @property
    def q(self):
        return numpy.array(self._q.value, dtype=int)

    @q.setter
    def q(self, q: List[int]):
        self._q.setValue(q)

    @property
    def a(self) -> float:
        return self._a.value()

    @a.setter
    def a(self, a: float):
        self._a.setValue(a)

    @property
    def map_range(self) -> float:
        return self._map_range.value()

    @map_range.setter
    def map_range(self, map_range: float):
        self._map_range.setValue(map_range)

    @property
    def detector(self):
        return self._detector.currentText()

    @detector.setter
    def detector(self, detector: str):
        self._detector.setCurrentText(detector)

    @property
    def units(self):
        return self._units.currentText()

    @units.setter
    def units(self, units: str):
        self._units.setCurrentText(units)

    @property
    def n(self):
        return numpy.array(self._n.value, dtype=int)

    @n.setter
    def n(self, n: List[int]):
        self._n.setValue(n)

    @property
    def map_shape(self):
        return numpy.array(self._map_shape.value, dtype=int)

    @map_shape.setter
    def map_shape(self, map_shape: List[int]):
        self._map_shape.setValue(map_shape)

    @property
    def energy(self) -> float:
        return self._energy.value()

    @energy.setter
    def energy(self, energy: float):
        self._energy.setValue(energy)

    def updatePlot(self, arr: numpy.ndarray, edges: numpy.ndarray):
        arr = numpy.nan_to_num(arr)

        self._projections = []

        if self._units.currentText().lower() == "poulsen":
            self.labels = [r"$q_{rock}$", r"$q_{\perp}$", r"$q_{||}$"]
        else:
            self.labels = ["h", "k", "l"]

        for idx, (i, j) in enumerate(self._idx):
            self._projections += [numpy.sum(arr, axis=idx)]
            xscale = (edges[i][-1] - edges[i][0]) / len(edges[i])
            yscale = (edges[j][-1] - edges[j][0]) / len(edges[j])
            self._plots[idx].addImage(
                self._projections[-1],
                scale=(xscale, yscale),
                origin=(edges[i][0], edges[j][0]),
                xlabel=self.labels[i],
                ylabel=self.labels[j],
            )

        self._edges = edges

        self._plotWidget.show()
        self._exportButton.show()

    def exportMaps(self):
        """
        Creates dictionary with maps information and exports it to a nexus file
        """
        entry = "entry"

        nx = {
            entry: {"@NX_class": "NXentry"},
            "@NX_class": "NXroot",
            "@default": "entry",
        }

        for idx, (j, i) in enumerate(self._idx):
            axis = [self._edges[i][:-1], self._edges[j][:-1]]
            nx["entry"][str(idx)] = create_nxdata_dict(
                self._projections[idx],
                self.labels[i] + "_" + self.labels[j],
                axis,
                axes_names=[self.labels[i], self.labels[j]],
            )

        filename = select_output_hdf5_file_with_dialog()
        if filename:
            dicttonx(nx, filename)



from __future__ import annotations

import numpy
import pytest
from ewoksorange.tests.conftest import qtapp  # noqa F811
from silx.gui import qt
from silx.gui.data.test.test_dataviewer import SignalListener
from silx.gui.widgets.RangeSlider import RangeSlider

from darfix.gui.utils.rangeSlider import RangeSliderWithSpinBox


@pytest.fixture
def widget(qtapp) -> RangeSliderWithSpinBox:  # noqa F811
    w = RangeSliderWithSpinBox()
    return w


@pytest.fixture
def min_spin_box(widget) -> qt.QDoubleSpinBox:
    return widget._RangeSliderWithSpinBox__spinBoxMin


@pytest.fixture
def max_spin_box(widget) -> qt.QDoubleSpinBox:
    return widget._RangeSliderWithSpinBox__spinBoxMax


@pytest.fixture
def slider(widget) -> RangeSlider:
    return widget._RangeSliderWithSpinBox__slider


def test_set_range_updates_spins_and_slider(widget, slider, min_spin_box, max_spin_box):
    vmin, vmax = -10.0, 10.0
    widget.setRange(vmin, vmax)

    # Check spin boxes
    assert min_spin_box.value() == pytest.approx(vmin)
    assert max_spin_box.value() == pytest.approx(vmax)

    # Check slider
    assert slider.getMinimum() == pytest.approx(vmin)
    assert slider.getMaximum() == pytest.approx(vmax)
    smin, smax = slider.getValues()
    assert smin == pytest.approx(vmin)
    assert smax == pytest.approx(vmax)


def test_min_spin_box(widget, min_spin_box, qtapp):  # noqa F811
    widget.setRange(-2.0, 2.0)
    listener = SignalListener()
    widget.sigValueChanged.connect(listener)

    min_spin_box.setValue(0.0)
    min_spin_box.editingFinished.emit()
    qtapp.processEvents()

    vmin, vmax = listener.arguments(0)
    numpy.testing.assert_allclose(listener.arguments(0), widget.getValues())
    assert vmin == pytest.approx(0.0)
    assert vmax == pytest.approx(2.0)

    min_spin_box.setValue(-20.0)
    min_spin_box.editingFinished.emit()
    qtapp.processEvents()

    vmin, vmax = listener.arguments(1)
    numpy.testing.assert_allclose(listener.arguments(1), widget.getValues())
    assert vmin == pytest.approx(-20.0)
    assert vmax == pytest.approx(2.0)
    assert widget.getRange()[0] == pytest.approx(-20.0)

    min_spin_box.setValue(20.0)
    min_spin_box.editingFinished.emit()
    qtapp.processEvents()

    vmin, vmax = listener.arguments(2)
    assert vmin == pytest.approx(2.0)  # max value is maxspin value
    assert vmax == pytest.approx(2.0)


def test_max_spin_box(widget, max_spin_box, qtapp):  # noqa F811
    widget.setRange(-2.0, 2.0)
    listener = SignalListener()
    widget.sigValueChanged.connect(listener)

    max_spin_box.setValue(1.0)
    max_spin_box.editingFinished.emit()
    qtapp.processEvents()

    vmin, vmax = listener.arguments(0)
    numpy.testing.assert_allclose(listener.arguments(0), widget.getValues())
    assert vmin == pytest.approx(-2.0)
    assert vmax == pytest.approx(1.0)

    max_spin_box.setValue(20.0)
    max_spin_box.editingFinished.emit()
    qtapp.processEvents()

    vmin, vmax = listener.arguments(1)
    numpy.testing.assert_allclose(listener.arguments(1), widget.getValues())
    assert vmin == pytest.approx(-2.0)
    assert vmax == pytest.approx(20.0)
    assert widget.getRange()[1] == pytest.approx(20.0)

    max_spin_box.setValue(-20.0)
    max_spin_box.editingFinished.emit()
    qtapp.processEvents()

    vmin, vmax = listener.arguments(2)
    assert vmin == pytest.approx(-2.0)  # max value is maxspin value
    assert vmax == pytest.approx(-2.0)  # min value is minspin value


def test_slider(widget, slider, min_spin_box, max_spin_box, qtapp):  # noqa F811
    widget.setRange(0.0, 10.0)
    listener = SignalListener()
    widget.sigValueChanged.connect(listener)

    slider.setValues(3.0, 7.0)
    qtapp.processEvents()

    smin, smax = listener.arguments(0)
    assert smin == pytest.approx(3.0)
    assert smax == pytest.approx(7.0)
    assert smin == pytest.approx(min_spin_box.value())
    assert smax == pytest.approx(max_spin_box.value())


def test_invalid_range_raises(widget):
    with pytest.raises(ValueError):
        widget.setRange(10.0, 5.0)



from __future__ import annotations

import numpy
import pytest
from ewoksorange.tests.conftest import qtapp  # noqa F811
from silx.gui.data.test.test_dataviewer import SignalListener

from darfix.core.dimension import AcquisitionDims
from darfix.core.dimension import Dimension
from darfix.gui.grainplot.dimensionRangeSlider2D import DimensionRangeSlider2D
from darfix.gui.grainplot.dimensionRangeSlider2D import _DimensionRangeSlider


@pytest.fixture
def widget(qtapp) -> DimensionRangeSlider2D:  # noqa F811
    return DimensionRangeSlider2D()


@pytest.fixture
def dims() -> AcquisitionDims:
    dims = AcquisitionDims()
    dims.add_dim(0, Dimension("A", 30, 0, 10))
    dims.add_dim(1, Dimension("B", 20, 0, 3))
    dims.add_dim(2, Dimension("C", 8, -10, -12))
    return dims


def test_set_dimensions(widget, dims):
    listener = SignalListener()

    widget.sigChanged.connect(listener)
    widget.setDimensions(dims)
    # Dim X is "A"
    assert widget.indexDimX() == 0
    numpy.testing.assert_allclose(widget.rangeDimX(), [0, 10])
    # Dim Y is "B"
    assert widget.indexDimY() == 1
    numpy.testing.assert_allclose(widget.rangeDimY(), [0, 3])
    assert (
        listener.callCount() == 0
    ), "`sigChanged` should not be called by `setDimensions`"


def test_switch_dimensions(widget, dims):

    listener = SignalListener()

    widget.sigChanged.connect(listener)
    widget.setDimensions(dims)
    sliders: tuple[_DimensionRangeSlider, _DimensionRangeSlider] = (
        widget._DimensionRangeSlider2D__sliders
    )

    sliders[0].setDimensionIndex(2)
    assert listener.callCount() == 1
    # Dim X is now "C"
    assert widget.indexDimX() == 2
    numpy.testing.assert_allclose(widget.rangeDimX(), [-12, -10])
    # Dim Y is still "B"
    assert widget.indexDimY() == 1
    numpy.testing.assert_allclose(widget.rangeDimY(), [0, 3])

    sliders[0].setDimensionIndex(1)
    assert listener.callCount() == 2
    # Dim X is now "B"
    assert widget.indexDimX() == 1
    numpy.testing.assert_allclose(widget.rangeDimX(), [0, 3])
    # Dim Y move to "A" as first slider is "B" (ANd we cannot have the same dimension for x and y)
    assert widget.indexDimY() == 0
    numpy.testing.assert_allclose(widget.rangeDimY(), [0, 10])



from __future__ import annotations

from typing import Tuple

from silx.gui import qt

from ..utils.vspacer import VSpacer


class ShiftInput(qt.QWidget):
    """
    Widget used to obtain the double parameters for the shift correction.
    """

    shiftChanged = qt.Signal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.findShiftB = qt.QPushButton("Find shift")
        self.abortShiftB = qt.QPushButton("Abort")
        # First dim is displayed on vertical axis, second on horizontal
        firstDimLabel = qt.QLabel("Vertical shift per frame (pixels)")
        secondDimLabel = qt.QLabel("Horizontal shift per frame (pixels)")
        self._firstDimSB = qt.QDoubleSpinBox()
        self._secondDimSB = qt.QDoubleSpinBox()

        for spinbox in (self._firstDimSB, self._secondDimSB):
            spinbox.setMinimum(-1000)
            spinbox.setMaximum(1000)
            spinbox.setButtonSymbols(qt.QAbstractSpinBox.ButtonSymbols.NoButtons)

        self._firstDimSB.editingFinished.connect(self.shiftChanged.emit)
        self._secondDimSB.editingFinished.connect(self.shiftChanged.emit)

        layout = qt.QGridLayout()

        layout.addWidget(self.findShiftB, 0, 0, 1, 1)
        layout.addWidget(self.abortShiftB, 0, 1, 1, 1)
        layout.addWidget(firstDimLabel, 1, 0)
        layout.addWidget(secondDimLabel, 2, 0)
        layout.addWidget(self._firstDimSB, 1, 1)
        layout.addWidget(self._secondDimSB, 2, 1)

        layout.addWidget(VSpacer())

        self.setLayout(layout)

    def getShift(self) -> Tuple[float, float]:
        return float(self._firstDimSB.value()), float(self._secondDimSB.value())

    def setShift(self, shift: Tuple[float, float]):
        first_dim_shift, second_dim_shift = shift
        self._firstDimSB.setValue(first_dim_shift)
        self._secondDimSB.setValue(second_dim_shift)



from __future__ import annotations

from ewokscore.missing_data import MISSING_DATA

from darfix.gui.zSumWidget import ZSumWidget
from darfix.tasks.zsum import ZSum
from orangecontrib.darfix.widgets.datasetWidgetBase import DatasetWidgetBase


class ZSumWidgetOW(DatasetWidgetBase, ewokstaskclass=ZSum):
    """
    Widget that compute and display the Z-sum of a dataset
    """

    name = "z sum"
    icon = "icons/zsum.svg"
    want_main_area = True
    want_control_area = False

    _ewoks_inputs_to_hide_from_orange = ("indices", "selected_axis")

    def __init__(self):
        super().__init__()

        self._widget = ZSumWidget(parent=self)
        self.mainArea.layout().addWidget(self._widget)
        # connect signal / slot
        self._widget.sigAxisChanged.connect(self._onAxisChanged)
        self._widget.sigResetFiltering.connect(self._onUncheckFiltering)

        self.datasetChanged.connect(self.onDatasetChanged)
        self.datasetChanged.connect(self._widget.setDataset)

    def onDatasetChanged(self) -> None:
        self.set_dynamic_input("selected_axis", None)
        self.execute_ewoks_task()

    def task_output_changed(self):
        z_sum = self.get_task_output_value("zsum", MISSING_DATA)
        if z_sum is not MISSING_DATA:
            self._widget.setZSum(z_sum)

    def _onAxisChanged(self, selectedAxis: int):
        self.set_default_input("indices", self._widget.indices)
        self.set_dynamic_input("selected_axis", selectedAxis)
        self.execute_ewoks_task_without_propagation()

    def _onUncheckFiltering(self):
        self.set_default_input("indices", None)
        self.set_dynamic_input("selected_axis", None)
        self.execute_ewoks_task_without_propagation()



from __future__ import annotations

from enum import Enum

from silx.gui import icons
from silx.gui import qt

from ...dtypes import AxisType
from ..utils.axis_type_combobox import AXIS_TYPES


class OriDistButtonIds(Enum):
    DATA = 0
    """Data of the orientation distribution histogram 2D"""
    COLOR_KEY = 1
    """ Data as a HSV color map"""


class _AxisTypeToolButton(qt.QToolButton):
    """Select axis from `AXIS_TYPES` with a menu"""

    sigChanged = qt.Signal()

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent=parent)
        self.setText("Axis")
        self.setToolTip("Select axis type")
        self.setPopupMode(qt.QToolButton.InstantPopup)

        self._actionGroup = qt.QActionGroup(self)
        self._actionGroup.setExclusive(True)

        for key in AXIS_TYPES.keys():
            action = qt.QAction(key, self)
            action.setCheckable(True)
            action.triggered.connect(self.sigChanged)
            self._actionGroup.addAction(action)
            self.addAction(action)

        self._actionGroup.actions()[0].setChecked(True)

    def getCurrentAxisType(self) -> AxisType:
        return AXIS_TYPES[self._actionGroup.checkedAction().text()]


class _NormalizationToolButton(qt.QToolButton):
    """Select normalization (Log or Linear) with a menu"""

    sigChanged = qt.Signal()

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent=parent)
        self.setText("Normalization")
        self.setToolTip("Select normalization for plot and contours")
        self.setPopupMode(qt.QToolButton.InstantPopup)

        self._actionGroup = qt.QActionGroup(self)
        self._actionGroup.setExclusive(True)

        for normalization in ["log", "linear"]:
            action = qt.QAction(normalization, self)
            action.setCheckable(True)
            action.triggered.connect(self.sigChanged)
            self._actionGroup.addAction(action)
            self.addAction(action)

        self._actionGroup.actions()[0].setChecked(True)

    def getNormalization(self) -> str:
        return self._actionGroup.checkedAction().text()


class _SliderToolButton(qt.QToolButton):
    """base class for a checkable toolbutton with a slider popup"""

    sigChanged = qt.Signal()

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent=parent)
        self.setPopupMode(qt.QToolButton.MenuButtonPopup)
        self.setCheckable(True)
        self.setChecked(True)
        self.toggled.connect(self.sigChanged)

        self._slider = qt.QSlider(qt.Qt.Orientation.Horizontal)
        self._slider.setRange(0, 255)
        self._slider.valueChanged.connect(self.sigChanged)

        self._description = qt.QLabel()

        w = qt.QWidget()
        w.setLayout(qt.QVBoxLayout())
        w.layout().addWidget(self._description)
        w.layout().addWidget(self._slider)

        # embed the slider into the menu
        action = qt.QWidgetAction(self)
        action.setDefaultWidget(w)
        self.addAction(action)

    def value(self) -> int:
        return self._slider.value()


class _ContourConfigToolButton(_SliderToolButton):
    """Configure the contours"""

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent=parent)
        self.setIcon(icons.getQIcon("darfix:gui/icons/contour"))
        self._slider.setRange(0, 30)
        self._slider.setValue(10)
        self.setToolTip("Contours count configuration. Disable contours if unchecked.")
        self._description.setText("Contours count")


class _DatatypeToolbutton(_SliderToolButton):
    """Select plot to show between data or color key. Configure transparency of color key"""

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent=parent)
        self.setIcon(icons.getQIcon("darfix:gui/icons/hsv"))
        self._slider.setRange(0, 255)
        self._slider.setValue(127)
        self.setToolTip(
            "Tune color key layer opacity. Show underlying data if unchecked."
        )
        self._description.setText("Tune color key layer opacity")



from __future__ import annotations

import numpy
from silx.gui import qt

from darfix.gui.shiftcorrection.shiftCorrectionWidget import ShiftCorrectionWidget
from darfix.tasks.shiftcorrection import ShiftCorrection
from orangecontrib.darfix.widgets.operationWidgetBase import OperationWidgetBase


class ShiftCorrectionWidgetOW(OperationWidgetBase, ewokstaskclass=ShiftCorrection):
    """
    Widget to make the shift correction of a dataset.
    """

    name = "shift correction"
    description = "A widget to perform shift correction"
    icon = "icons/shift_correction.svg"

    _ewoks_inputs_to_hide_from_orange = (
        "shift",
        "selected_axis",
        "selected_index",
        "copy_dataset",
    )

    def __init__(self):
        super().__init__()
        qt.QLocale.setDefault(qt.QLocale("en_US"))
        self.datasetChanged.connect(self.mainWidget.setDataset)
        self.sigOperationApplied.connect(self.mainWidget.updateDataset)

    def createMainWidget(self, default_inputs: dict) -> ShiftCorrectionWidget:
        widget = ShiftCorrectionWidget()
        widget.setCorrectionInputs(
            default_inputs.get("shift", (0, 0)),
            default_inputs.get("selected_axis", None),
        )
        return widget

    def saveMainWidget(self) -> dict:
        inputs = self.mainWidget.getCorrectionInputs()
        for key, value in inputs.items():
            if isinstance(value, numpy.ndarray):
                value = value.tolist()
            self.set_default_input(key, value)

    @property
    def mainWidget(self) -> ShiftCorrectionWidget:
        return super().mainWidget



from __future__ import annotations

from darfix.core.noiseremoval import clean_operation_dict
from darfix.gui.noiseremoval.noise_removal_widget import NoiseRemovalWidget
from darfix.tasks.noiseremoval import NoiseRemoval
from orangecontrib.darfix.widgets.operationWidgetBase import OperationWidgetBase


class NoiseRemovalWidgetOW(OperationWidgetBase, ewokstaskclass=NoiseRemoval):
    name = "noise removal"
    description = "A widget to perform various noise removal operations"
    icon = "icons/noise_removal.png"

    _ewoks_inputs_to_hide_from_orange = ("operations", "copy_dataset")

    def __init__(self):
        super().__init__()
        self.datasetChanged.connect(self.mainWidget.setDataset)
        self.sigOperationApplied.connect(self.mainWidget.setAppliedState)

    def createMainWidget(self, inputs: dict) -> NoiseRemovalWidget:
        widget = NoiseRemovalWidget()
        incomingOperations = inputs.get("operations", [])
        widget.setDefaultParameters(incomingOperations)
        return widget

    def saveMainWidget(self):
        self.set_default_input(
            "operations",
            [
                clean_operation_dict(operation)
                for operation in self.mainWidget.getOperationList()
            ],
        )

    @property
    def mainWidget(self) -> NoiseRemovalWidget:
        return super().mainWidget



import importlib.metadata
import sys
from argparse import ArgumentParser

from silx import config

try:
    from ewoksorange.canvas.main import arg_parser
    from ewoksorange.canvas.main import main as ewoksorange_main
except ImportError as e:
    error_msg = f"ERROR: {e.msg}.\n"
    error_msg += "To use `darfix` command, please use the full installation of darfix:\npip install darfix[full]\n"
    sys.stdout.write(error_msg)
    exit()


def main(argv=None):

    config._MPL_TIGHT_LAYOUT = True

    parser = ArgumentParser(parents=[arg_parser()], add_help=False)

    parser.add_argument(
        "--version",
        action="store_true",
        help="Display version",
    )

    parser.add_argument(
        "--use-opengl-plot",
        action="store_true",
        help="Use opengl as default backend for all plots. Faster but some limitations : https://www.silx.org/doc/silx/2.2.1/troubleshooting.html",
    )

    if argv is None:
        argv = sys.argv
    options, _ = parser.parse_known_args(argv[1:])

    if options.use_opengl_plot:
        config.DEFAULT_PLOT_BACKEND = "gl"
        argv.pop(argv.index("--use-opengl-plot"))

    if options.version:
        print(f"Darfix version: {importlib.metadata.version('darfix')}")
        return
    if ewoksorange_main is None:
        raise ImportError("Install darfix[full] to use the Orange canvas.")
    ewoksorange_main()


if __name__ == "__main__":
    sys.exit(main())
